# Copyright (c) Microsoft Corporation.
# Licensed under the MIT license.

import numpy as np
from pycuda import gpuarray
import pycuda.driver as cuda

from svirl import config as cfg

from svirl.storage import GArray 

class Vars(object):
    """This class contains setters and getters for solution variables
    order parameter and vector potential"""

    def __init__(self, Par, mesh):
        self.par = Par
        self.mesh = mesh

        # Persistent storage: Primary variables that are solved for
        self._psi = None
        self._vp = None # TODO: this will be renamed to self._a eventually

        # Temp storage: Stored on cells, nodes, and edges. 
        # Used in observables and other classes and fairly general purpose
        self._tmp_node_var = None
        self._tmp_edge_var = None
        self._tmp_cell_var = None

        # Temp Storage: Allocated only for reductions
        self._tmp_psi_real = None
        self._tmp_A_real = None

        # copied from variables/parameters.py
        self.solveA = np.bool(not np.isposinf(cfg.gl_parameter))

        if cfg.order_parameter == 'random': 
            self.order_parameter = 1.0
            self.randomize_order_parameter(level = cfg.random_level, seed = cfg.random_seed)
        else:
            self.order_parameter = cfg.order_parameter   # set order parameter manually

        # vector potential is set up here instead of its setter because 
        # we don't plan on supporting setter for it
        shapes = [(cfg.Nxa, cfg.Nya), (cfg.Nxb, cfg.Nyb)]
        self._vp = GArray(shape = shapes, dtype = cfg.dtype)


    def __del__(self):
        pass


    @property
    def order_parameter(self):
        self._psi.sync()
        psi = self._psi.get_h().copy()
        return psi


    @order_parameter.setter
    def order_parameter(self, order_parameter):
        if isinstance(order_parameter, (np.complexfloating, complex, np.floating, float, np.integer, int)):
            order_parameter = cfg.dtype_complex(order_parameter) * np.ones((cfg.Nx, cfg.Ny), cfg.dtype_complex)
        assert order_parameter.shape == (cfg.Nx, cfg.Ny)

        if self._psi is None:
            self._psi = GArray(like = order_parameter)
        else:
            self._psi.set_h(order_parameter)

        self.set_order_parameter_to_zero_outside_material()
        self._psi.sync()


    def order_parameter_h(self):
        return self._psi.get_d_obj() 


    def set_order_parameter_to_zero_outside_material(self):

        if self._psi is None or not self.mesh.have_material_tiling():
            return

        mt_at_nodes = self.mesh._get_material_tiling_at_nodes()

        psi = self._psi.get_h()
        psi[~mt_at_nodes] = 0.0

        self._psi.need_htod_sync()
        self._psi.sync()


    def randomize_order_parameter(self, level=1.0, seed=None):
        """Randomizes order parameter:
            absolute value *= 1 - level*rand
            phase += level*pi*(2.0*rand()-1.0),
            where rand is uniformly distributed in [0, 1]
        """
        assert 0.0 <= level <= 1.0

        self._psi.sync()

        if seed is not None:
            np.random.seed(seed)

        data = (1.0 - level*np.random.rand(cfg.N)) * np.exp(level * 1.0j*np.pi*(2.0*np.random.rand(cfg.N) - 1.0))

        self._psi.set_h(data)
        self._psi.sync()


    @property
    def vector_potential(self):
        if self._vp is None:
            return (np.zeros((cfg.Nxa, cfg.Nya), dtype=cfg.dtype),
                    np.zeros((cfg.Nxb, cfg.Nyb), dtype=cfg.dtype))

        self._vp.sync()
        return self._vp.get_vec_h() 


    @vector_potential.setter
    def vector_potential(self, vector_potential):

        a, b = vector_potential
        self._vp.set_vec_h(a, b)
        self._vp.sync()


    def vector_potential_h(self):
        if self._vp is not None:
            return self._vp.get_d_obj()

        return np.uintp(0)

    #--------------------------
    # temporary arrays
    #--------------------------

    def _tmp_node_var_h(self):
        if self._tmp_node_var is None:
            self._tmp_node_var = GArray(like = self._psi)

        return self._tmp_node_var.get_d_obj() 


    def _tmp_edge_var_h(self):
        if self._tmp_edge_var is None:
            shapes = [(cfg.Nxa, cfg.Nya), (cfg.Nxb, cfg.Nyb)]
            self._tmp_edge_var = GArray(shape = shapes, dtype = cfg.dtype)

        return self._tmp_edge_var.get_d_obj()


    def _tmp_cell_var_h(self):
        if self._tmp_cell_var is None:
            self._tmp_cell_var = GArray(shape = (cfg.Nxc, cfg.Nyc), 
                                          dtype = cfg.dtype)

        return self._tmp_cell_var.get_d_obj()


    def _tmp_psi_real_h(self):
        if self._tmp_psi_real is not None:
            return self._tmp_psi_real.get_d_obj()

        return np.uintp(0)


    def _tmp_A_real_h(self):
        if self._tmp_A_real is not None:
            return self._tmp_A_real.get_d_obj()

        return np.uintp(0)

    def _alloc_free_temporary_gpu_storage(self, action):
        assert action in ['alloc', 'free']

        if action == 'alloc':
            if self._tmp_psi_real is None:
                self._tmp_psi_real = GArray(self.par.grid_size, on =
                        GArray.on_device, dtype = cfg.dtype)
            if self._tmp_A_real is None and self.solveA:
                self._tmp_A_real = GArray(self.par.grid_size_A, on =
                        GArray.on_device, dtype = cfg.dtype)
        else:
            if self._tmp_psi_real is not None:
                self._tmp_psi_real.free()
                self._tmp_psi_real = None
            if self._tmp_A_real is not None:
                self._tmp_A_real.free()
                self._tmp_A_real = None





# TODO: fix resolution issue (for larger systems)
# TODO: implement arbitrary array
# TODO: implement movies
#         e.g. using https://github.com/kkroening/ffmpeg-python

import numpy as np
import copy

import matplotlib
matplotlib.use('Agg')
import matplotlib.pyplot as plt

from PIL import Image

import cmocean


def __get_quantity(svirl, type):
    def __nph(phase):
        return np.mod(phase/(2.0*np.pi) + 0.5, 1.0) - 0.5
    
    if   type == 'material_tiling':
        quantity = svirl.mesh.material_tiling.astype(svirl.cfg.dtype)
    if   type == 'linear_coefficient':
        quantity = svirl.params.linear_coefficient
    elif type == 'superfluid_density':
        quantity = svirl.observables.superfluid_density
    elif type == 'order_parameter_phase':
        quantity = np.angle(svirl.vars.order_parameter)/(2.0*np.pi)
    elif type == 'op_fv_phase':
        quantity = __nph(np.angle(svirl.vars.order_parameter) - svirl.params.fixed_vortices.fixed_vortices_phase)
    elif type == 'magnetic_field':
        quantity = svirl.observables.magnetic_field - svirl.params.homogeneous_external_field
    elif type == 'current_density':
        jx, jy = svirl.observables.current_density
        quantity = svirl.mesh.interpolate_ab_array_to_c_array_abs(jx, jy)
    elif type == 'current_density_x':
        quantity, _ = svirl.observables.current_density
    elif type == 'current_density_y':
        _, quantity, = svirl.observables.current_density
    elif type == 'supercurrent_density':
        jsx, jsy = svirl.observables.supercurrent_density
        quantity = svirl.mesh.interpolate_ab_array_to_c_array_abs(jsx, jsy)
    elif type == 'supercurrent_density_x':
        quantity, _ = svirl.observables.supercurrent_density
    elif type == 'supercurrent_density_y':
        _, quantity, = svirl.observables.supercurrent_density    
    elif type == 'normalcurrent_density':
        jsx, jsy = svirl.observables.normalcurrent_density
        quantity = svirl.mesh.interpolate_ab_array_to_c_array_abs(jsx, jsy)
    elif type == 'normalcurrent_density_x':
        quantity, _ = svirl.observables.normalcurrent_density
    elif type == 'normalcurrent_density_y':
        _, quantity, = svirl.observables.normalcurrent_density    
    elif type == 'vector_potential_x':
        quantity, _ = svirl.vars.vector_potential
    elif type == 'vector_potential_y':
        _, quantity = svirl.vars.vector_potential
    elif type == 'irregular_vector_potential_x':
        quantity, _ = svirl.params.fixed_vortices.irregular_vector_potential
    elif type == 'irregular_vector_potential_y':
        _, quantity = svirl.params.fixed_vortices.irregular_vector_potential
    elif type == 'fixed_vortices_phase':
        quantity = __nph(svirl.params.fixed_vortices.fixed_vortices_phase)
    
    return np.flipud(quantity.T)


def __get_cmap(type):
    return {
        'material_tiling'                     : plt.get_cmap('binary'),
        'linear_coefficient'                  : cmocean.cm.haline,
        'superfluid_density'                  : plt.get_cmap('nipy_spectral'),
        'order_parameter_phase'               : cmocean.cm.curl,                          # plt.get_cmap('hsv')
        'op_fv_phase'                         : cmocean.cm.curl,
        'magnetic_field'                      : cmocean.cm.balance,                       # plt.get_cmap('PiYG_r')
        'current_density'                     : cmocean.cm.dense,                         # plt.get_cmap('gnuplot2')
        'current_density_x'                   : cmocean.cm.delta,                         # plt.get_cmap('RdBu_r')
        'current_density_y'                   : cmocean.cm.delta,
        'supercurrent_density'                : cmocean.cm.dense,
        'supercurrent_density_x'              : cmocean.cm.delta,
        'supercurrent_density_y'              : cmocean.cm.delta,
        'normalcurrent_density'               : cmocean.cm.dense,
        'normalcurrent_density_x'             : cmocean.cm.delta,
        'normalcurrent_density_y'             : cmocean.cm.delta,
        'vector_potential_x'                  : cmocean.cm.tarn,                          # plt.get_cmap('seismic')
        'vector_potential_y'                  : cmocean.cm.tarn,
        'irregular_vector_potential_x'        : cmocean.cm.tarn,
        'irregular_vector_potential_y'        : cmocean.cm.tarn,
        'fixed_vortices_phase'                : cmocean.cm.curl,
    }[type]


def __expand_types(types):
    if not isinstance(types, tuple): types = (types, )
    expanders = {
        'order_parameter'               : ['superfluid_density',           'order_parameter_phase'        ],
        'current_density_xy'            : ['current_density_x',            'current_density_y'            ],
        'supercurrent_density_xy'       : ['supercurrent_density_x',       'supercurrent_density_y'       ],
        'normalcurrent_density_xy'      : ['normalcurrent_density_x',      'normalcurrent_density_y'      ],
        'vector_potential_xy'           : ['vector_potential_x',           'vector_potential_y'           ],
        'irregular_vector_potential_xy' : ['irregular_vector_potential_x', 'irregular_vector_potential_xy'],
    }
    expanded_types, indices = [], []
    for i, type in enumerate(types):
        if type in expanders:
            for t in expanders[type]:
                expanded_types.append(t)
                indices.append(i)
        else:
            expanded_types.append(type)
            indices.append(i)
    return expanded_types, indices


__sub_width_px, __sub_height_px = None, None
def __plot_one(svirl, type='order_parameter', **kwargs):
    """Plots set of (2D) quantities to current gca"""
    
    global __sub_width_px, __sub_height_px
    
    if 'title' in kwargs:
        title = kwargs['title']
    else:
        title = {
            'material_tiling'                     : 'material',
            'linear_coefficient'                  : '$\\epsilon$',
            'superfluid_density'                  : '$|\\psi|^2$',
            'order_parameter_phase'               : '$\\mathrm{arg}(\\psi)/2\\pi$',
            'op_fv_phase'                         : '$[\\mathrm{arg}(\\psi) - \\int d\\mathbf{l}\\,\\mathbf{A}_{\\mathrm{i}}] \\,/\\, 2\\pi$',
            'magnetic_field'                      : '$B(x,y) - H$',
            'current_density'                     : '$|\\mathbf{j}|$',
            'current_density_x'                   : '$j_x$',
            'current_density_y'                   : '$j_y$',
            'supercurrent_density'                : '$|\\mathbf{j}_\\mathrm{s}|$',
            'supercurrent_density_x'              : '$j_{\\mathrm{s},x}$',
            'supercurrent_density_y'              : '$j_{\\mathrm{s},y}$',
            'normalcurrent_density'               : '$|\\mathbf{j}_\\mathrm{n}|$',
            'normalcurrent_density_x'             : '$j_{\\mathrm{n},x}$',
            'normalcurrent_density_y'             : '$j_{\\mathrm{n},y}$',
            'vector_potential_x'                  : '$A_x$',
            'vector_potential_y'                  : '$A_y$',
            'irregular_vector_potential_x'        : '$A_{\\mathrm{i},x}$',
            'irregular_vector_potential_y'        : '$A_{\\mathrm{i},y}$',
            'fixed_vortices_phase'                : '$\\int d\\mathbf{l}\\,\\mathbf{A}_{\\mathrm{i}} \\,/\\, 2\\pi$'
        }[type]

    grid_type = {
        'material_tiling'                     : 'c',
        'linear_coefficient'                  : 'psi',
        'superfluid_density'                  : 'psi',
        'order_parameter_phase'               : 'psi',
        'op_fv_phase'                         : 'psi',
        'magnetic_field'                      : 'c',
        'current_density'                     : 'c',
        'current_density_x'                   : 'a',
        'current_density_y'                   : 'b',
        'supercurrent_density'                : 'c',
        'supercurrent_density_x'              : 'a',
        'supercurrent_density_y'              : 'b',
        'normalcurrent_density'               : 'c',
        'normalcurrent_density_x'             : 'a',
        'normalcurrent_density_y'             : 'b',
        'vector_potential_x'                  : 'a',
        'vector_potential_y'                  : 'b',
        'irregular_vector_potential_x'        : 'a',
        'irregular_vector_potential_y'        : 'b',
        'fixed_vortices_phase'                : 'psi',
    }[type]
    
    if 'extent' in kwargs:
        extent = kwargs['extent']
    else:
        extent = {
            'psi': [-0.5*svirl.cfg.dx,  svirl.cfg.Lx+0.5*svirl.cfg.dx,  -0.5*svirl.cfg.dy,  svirl.cfg.Ly+0.5*svirl.cfg.dy],
            'a':   [ 0.0          ,  svirl.cfg.Lx              ,  -0.5*svirl.cfg.dy,  svirl.cfg.Ly+0.5*svirl.cfg.dy],
            'b':   [-0.5*svirl.cfg.dx,  svirl.cfg.Lx+0.5*svirl.cfg.dx,   0.0          ,  svirl.cfg.Ly              ],
            'c':   [ 0.0          ,  svirl.cfg.Lx              ,   0.0          ,  svirl.cfg.Ly              ],
        }[grid_type]
    
    __sub_width_px  = (extent[1] - extent[0]) / svirl.cfg.dx
    __sub_height_px = (extent[3] - extent[2]) / svirl.cfg.dy
    
    quantity = __get_quantity(svirl, type)
    
    if 'cmap' in kwargs:
        cmap = kwargs['cmap']
    else:
        cmap = __get_cmap(type)
    
    if 'balance_zero' in kwargs:
        balance_zero = kwargs['balance_zero']
    else:
        balance_zero = type in [
            'magnetic_field', 
            'current_density_x',            'current_density_y', 
            'supercurrent_density_x',       'supercurrent_density_y',
            'normalcurrent_density_x',      'normalcurrent_density_y',
            'vector_potential_x',           'vector_potential_y',
            'irregular_vector_potential_x', 'irregular_vector_potential_y',
        ]
    
    if 'interpolation' in kwargs:
        interpolation = kwargs['interpolation']
    else:
        interpolation = {
            'material_tiling'                     : 'nearest',
            'linear_coefficient'                  : 'spline16',
            'superfluid_density'                  : 'bilinear',
            'order_parameter_phase'               : 'spline16',
            'op_fv_phase'                         : 'spline16',
            'magnetic_field'                      : 'spline16',
            'current_density'                     : 'spline16',
            'current_density_x'                   : 'spline16',
            'current_density_y'                   : 'spline16',
            'supercurrent_density'                : 'spline16',
            'supercurrent_density_x'              : 'spline16',
            'supercurrent_density_y'              : 'spline16',
            'normalcurrent_density'               : 'spline16',
            'normalcurrent_density_x'             : 'spline16',
            'normalcurrent_density_y'             : 'spline16',
            'vector_potential_x'                  : 'spline16',
            'vector_potential_y'                  : 'spline16',
            'irregular_vector_potential_x'        : 'spline16',
            'irregular_vector_potential_y'        : 'spline16',
            'fixed_vortices_phase'                : 'spline16',
        }[type]
    
    if 'show_vortices' in kwargs:
        show_vortices = kwargs['show_vortices']
    else:
        show_vortices = False
    
    plt.imshow(
        quantity, 
        cmap = cmap, 
        interpolation = interpolation, 
        extent = extent,
        aspect = 'equal',
    )
    
    if show_vortices:
        vx, vy, vv = svirl.vortex_detector.vortices
        if vx.size > 0:
            plt.plot(vx[vv>0], vy[vv>0], 'o', mec='w', mfc='None', markersize=10, linewidth=0.5)
            plt.plot(vx[vv<0], vy[vv<0], 'x', mec='w', mfc='None', markersize=10, linewidth=0.5)
    
    plt.xlabel(kwargs['xlabel'] if 'xlabel' in kwargs else '$x$ [$\\xi$]')
    plt.ylabel(kwargs['ylabel'] if 'ylabel' in kwargs else '$y$ [$\\xi$]')
    cb = plt.colorbar()
    cb.set_label(title)
    if 'clim' in kwargs and kwargs['clim'] is not None:
        plt.clim(kwargs['clim'][0], kwargs['clim'][1])
    elif balance_zero:
        qmin, qmax = np.nanmin(quantity), np.nanmax(quantity)
        if np.abs(qmax) > np.abs(qmin):
            plt.clim(-qmax, qmax)
        else:
            plt.clim(qmin, -qmin)
    
    plt.xlim([extent[0], extent[1]])
    plt.ylim([extent[2], extent[3]])


def plot(svirl, types='order_parameter', **kwargs):
    expanded_types, expanded_indices = __expand_types(types)

    magnification = float(kwargs['magnification']) if 'magnification' in kwargs else 1.0
    dpi = float(kwargs['dpi']) if 'dpi' in kwargs else 100.0
    
    n = len(expanded_types)
    nrows = int(np.floor(np.sqrt(n)))
    ncols = int(np.ceil(float(n)/float(nrows)))
    
    if 'font_family' in kwargs: plt.rcParams.update({'font.family': kwargs['font_family']})
    if 'font_weight' in kwargs: plt.rcParams.update({'font.weight': kwargs['font_weight']})
    if 'font_size' in kwargs: plt.rcParams.update({'font.size': kwargs['font_size']})
    
    fig, axs = plt.subplots(nrows, ncols, sharex=False, sharey=False)
    axs = np.ravel([axs])
    plt.tight_layout(
        pad = kwargs['tight_layout_pad'] if 'tight_layout_pad' in kwargs else 1.04,
        w_pad = kwargs['tight_layout_w_pad'] if 'tight_layout_w_pad' in kwargs else None,
        h_pad = kwargs['tight_layout_h_pad'] if 'tight_layout_h_pad' in kwargs else None)
    
    total_w_in, total_h_in = [], []
    for j in range(nrows):
        for i in range(ncols):
            k = j*ncols + i
            if k >= len(expanded_types): continue
            _type = expanded_types[k]
            
            kwargs_one = copy.deepcopy(kwargs)
            
            for a in ['title', 'extent', 'cmap', 'balance_zero', 'interpolation', 'show_vortices', 'xlabel', 'ylabel', 'clim']:
                if a in kwargs and isinstance(kwargs[a], tuple):
                    kwargs_one[a] = kwargs[a][expanded_indices[k]]
            
            ax = axs[k]
            plt.sca(ax)
            
            __plot_one(svirl, _type, **kwargs_one)
            
            x_fr, y_fr, w_fr, h_fr = ax.get_position().bounds
            w_in, h_in = float(__sub_width_px)*magnification/dpi, float(__sub_height_px)*magnification/dpi
            total_w_in.append(w_in/w_fr);  total_h_in.append(h_in/h_fr)
    
    fig.set_size_inches(np.mean(total_w_in), np.mean(total_h_in))
    
    if 'suptitle' in kwargs:
        fig.suptitle(kwargs['suptitle'])


def save(svirl, filename, types='order_parameter', **kwargs):
    """Plots and saves set of (2D) quantities to file"""
    plt.close('all')
    plot(svirl, types, **kwargs)
    plt.savefig(
        filename, 
        bbox_inches = kwargs['bbox_inches'] if 'bbox_inches' in kwargs else None,
#         dpi = kwargs['dpi'] if 'dpi' in kwargs else 100,
    )


def __plotsimple(svirl, type, **kwargs):
    quantity = __get_quantity(svirl, type)
    
    if 'cmap' in kwargs:
        cmap = kwargs['cmap']
    else:
        cmap = __get_cmap(type)

    min, max = np.min(quantity), np.max(quantity)
    if not np.isclose(max, min):
        quantity = (quantity - min) / (max - min)
    else:
        quantity *= 0.0
    
    return Image.fromarray(np.uint8(cmap(quantity)*255))


def savesimple(svirl, filename, types='superfluid_density', **kwargs):
    expanded_types, _ = __expand_types(types)
    
    images = []
    max_w, max_h = 0, 0
    for type in expanded_types:
        im = __plotsimple(svirl, type, **kwargs)
        w, h = im.size
        max_w, max_h = max(max_w, w), max(max_h, h)
        images.append(im)
    
    n = len(expanded_types)
    nrows = int(np.floor(np.sqrt(n)))
    ncols = int(np.ceil(float(n)/float(nrows)))
    
    border, spacing = 2, 4
    
    combined_im = Image.new(
        'RGB', 
        (
            max_w*ncols + spacing*(ncols-1) + 2*border, 
            max_h*nrows + spacing*(nrows-1) + 2*border
        ),
        (255, 255, 255)
    )
    
    for y in range(nrows):
        for x in range(ncols):
            i = x + y*ncols
            if i >= len(images): break
            combined_im.paste(
                images[i], (
                    border + x*(max_w + spacing), 
                    border + y*(max_h + spacing)
                )
            )
    
    combined_im.save(filename)




# Copyright (c) Microsoft Corporation.
# Licensed under the MIT license.

import os
from pycuda import compiler as cuda_compiler 
import pycuda.driver as cuda
import numpy as np

from svirl import config as cfg
from svirl.parallel.utils import Utils

class Startup(object):
    """This class contains infratructure related to the startup mechanism 
    for CUDA. The block size is specified in this class.
    """

    def __init__(self):

        self.__device_inited = False

        assert isinstance(cfg.device_id, (np.integer, int)) and cfg.device_id >= 0

        self.__device_id = cfg.device_id
        self.__init_device()
        
        dtype_c = {np.float32: 'float', 
                   np.float64: 'double'}[cfg.dtype]
        
        dtype_complex_c = {np.complex64: 'pycuda::complex<float>', 
                           np.complex128: 'pycuda::complex<double>'}[cfg.dtype_complex]

        # If gl_parameter is changed, this needs to change as well 
        solveA = np.bool(not np.isposinf(cfg.gl_parameter))  
        if solveA:
            self.reduction_vector_length = 17
        else: 
            self.reduction_vector_length = 5
        
        # PyCUDA supports compilation of one cuda file, so files are
        # concatenated and thus order of files is important
        cuda_files_parallel = ['common.h', 'block_reduction.h', 'reduction.h', 'utils.h']
        cuda_files_solvers_post = ['observables.h', 'td.h', 'cg.h']
        
        cuda_src_files = cuda_files_parallel + cuda_files_solvers_post 

        this_dir = os.path.dirname(__file__)
        root_dir = os.path.abspath(os.path.join(this_dir, '..'))
        
        cuda_template = ''
        for ifile in cuda_src_files:
            with open(os.path.abspath(os.path.join(root_dir, 'cuda', ifile)), 'r') as f:
                cuda_template += f.read() + '\n'
        
        self.cuda_template_dict = {
            'real': dtype_c,
            'complex': dtype_complex_c,
            'Nx': cfg.Nx,  'Ny': cfg.Ny,
            'dx': cfg.dx,  'dy': cfg.dy,
            'reduction_vector_length': self.reduction_vector_length,
        } 

        cuda_code = cuda_template % self.cuda_template_dict 
        self._cuda_module = cuda_compiler.SourceModule(cuda_code, options=['-std=c++11'])

        self.block_size = 128
        self.grid_size = Utils.intceil(cfg.N, self.block_size)
        self.grid_size_A = Utils.intceil(cfg.Nab, self.block_size) 
        

    def __del__(self):
        self.__cuda_context.pop()


    def get_function(self, function_name):
        """Lookup and return the reference to the funtion if it is found,
        else raise an exception
        """
        try:
            return self._cuda_module.get_function(function_name)
        except:
            raise ValueError("\n Function name %s not found" \
                    % (function_name))


    ## Private methods

    def __init_device(self):
        if not self.__device_inited:
            cuda.init()  # TODO: driver needs to be inited only once, so this should be separate
            self.cuda_compute_capability = cuda.Device(self.__device_id).compute_capability()
            self.__cuda_context = cuda.Device(self.__device_id).make_context()

            self.__device_inited = True



# Copyright (c) Microsoft Corporation.
# Licensed under the MIT license.

import numpy as np
from pycuda import gpuarray

from svirl import config as cfg

from svirl import parallel  as GLPar 
from svirl import mesh      as GLMesh
from svirl import vars      as GLVars
from svirl import observables as GLObs
from svirl import solvers as GLSolvers


class GLSolver(object):
    """2D Ginzburg-Landau BDF solver
    
    Main features:
    - GPU (CUDA) implementation
    - time-dependence
    - finite GL parameter
    - user-defined material domain for order potential
    - fixed vortices can be added by using irregular vector potential
    - vortex detector
    - observables such as GL free energy, current and supercurrent densities, and induced magnetic field

    """

    def __init__(self,
        Nx = None, dx = None, Lx = None,                                                  # geometry
        Ny = None, dy = None, Ly = None, 
        
        Nt = None, dt = None, T = None,                                                   # parameters for order parameter iterator
        NtA = None, dtA = None, TA = None,                                                # parameters for order vector potential
        
        material_tiling = None,
        order_parameter = 'random',
        random_seed = None,
        random_level = 1.0,
        
        gl_parameter = np.inf,                                                            # = lambda/xi, GL parameter; if None, np.nan, or np.inf then solveA = False
        normal_conductivity = 1.0,                                                        # normal-state conductivity 
        linear_coefficient = 1.0,                                                         # linear coefficient in GL equation

        homogeneous_external_field = 0.0,
        external_field = 0.0,
        
        fixed_vortices = None,
        fixed_vortices_correction = 'cell centers',
        phase_lock_radius = None,
        
        device_id = 0,
        dtype = np.float64,
        
        stop_criterion_order_parameter = 1e-6,
        stop_criterion_vector_potential = 1e-6,
        
        order_parameter_Langevin_coefficient = 0.0,
        vector_potential_Langevin_coefficient = 0.0,

        convergence_rtol = 1e-6,  # relative tolerance for convergence
    ):

        self.dtypes = (np.float32, np.float64)

        cfg.device_id = device_id
        
        assert dtype in self.dtypes
        cfg.dtype = dtype
        cfg.dtype_complex = {np.float32: np.complex64, np.float64: np.complex128}[cfg.dtype]
        
        if   Nx is not None and Lx is not None and dx is None:
            dx = float(Lx)/(Nx-1)
        elif Nx is not None and Lx is None and dx is not None:
            Lx = float(dx)*(Nx-1)
        elif Nx is None and Lx is not None and dx is not None:
            Nx = int(np.round(Lx/dx)+1)
        elif Nx is not None and Lx is not None and dx is not None:
            assert np.isclose(Lx, dx*(Nx-1))
        else:
            raise 'Two out of three Nx, Lx, dx must be defined'
        
        if   Ny is not None and Ly is not None and dy is None:
            dy = float(Ly)/(Ny-1)
        elif Ny is not None and Ly is None and dy is not None:
            Ly = float(dy)*(Ny-1)
        elif Ny is None and Ly is not None and dy is not None:
            Ny = int(np.round(Ly/dy)+1)
        elif Ny is not None and Ly is not None and dy is not None:
            assert np.isclose(Ly, dy*(Ny-1))
        else:
            raise 'Two out of three Ny, Ly, and dy must be defined'
        
        assert isinstance(Lx, (np.floating, float, np.integer, int)) and Lx>0.0
        assert isinstance(Ly, (np.floating, float, np.integer, int)) and Ly>0.0
        assert isinstance(Nx, (np.integer, int)) and Nx>=4
        assert isinstance(Ny, (np.integer, int)) and Ny>=4
        
        cfg.Nx, cfg.Ny = np.int32(Nx),   np.int32(Ny)
        cfg.Lx, cfg.Ly = cfg.dtype(Lx), cfg.dtype(Ly)
        cfg.dx, cfg.dy = cfg.dtype(dx), cfg.dtype(dy)
        
        cfg.N = cfg.Nx*cfg.Ny
        
        # number of centers of horizontal edges excluding boundaries
        cfg.Nxa, cfg.Nya = cfg.Nx-1, cfg.Ny                 

        # number of centers of vertical edges excluding boundaries
        cfg.Nxb, cfg.Nyb = cfg.Nx,   cfg.Ny-1  

        cfg.Na, cfg.Nb = cfg.Nxa*cfg.Nya, cfg.Nxb*cfg.Nyb
        cfg.Nab = cfg.Na + cfg.Nb
        
        # number of cells
        cfg.Nxc, cfg.Nyc = cfg.Nx-1, cfg.Ny-1  
        cfg.Nc = cfg.Nxc*cfg.Nyc
        
        cfg.idx, cfg.idy = 1.0/cfg.dx, 1.0/cfg.dy
        cfg.idx2, cfg.idy2 = cfg.idx*cfg.idx, cfg.idy*cfg.idy

        cfg.idx2, cfg.idy2, cfg.idxy = cfg.idx*cfg.idx, cfg.idy*cfg.idy, cfg.idx*cfg.idy
        cfg.j_dx, cfg.j_dy = 1.0j*cfg.dx, 1.0j*cfg.dy

        cfg.material_tiling = material_tiling
        cfg.order_parameter = order_parameter
        cfg.random_seed = random_seed
        cfg.random_level = random_level

        cfg.gl_parameter = gl_parameter
        cfg.linear_coefficient = linear_coefficient
        cfg.normal_conductivity = normal_conductivity

        cfg.homogeneous_external_field = homogeneous_external_field
        cfg.external_field = external_field

        cfg.fixed_vortices = fixed_vortices
        cfg.fixed_vortices_correction = fixed_vortices_correction
        cfg.phase_lock_radius = phase_lock_radius

        cfg.order_parameter_Langevin_coefficient = order_parameter_Langevin_coefficient
        cfg.vector_potential_Langevin_coefficient = vector_potential_Langevin_coefficient

        cfg.Nt, cfg.dt, cfg.T = None, None, None
        cfg.NtA, cfg.dtA, cfg.TA = None, None, None

        assert isinstance(stop_criterion_order_parameter, (np.floating, float, np.integer, int)) and stop_criterion_order_parameter > 0.0
        cfg.stop_criterion_order_parameter = cfg.dtype(stop_criterion_order_parameter)
        
        assert isinstance(stop_criterion_vector_potential, (np.floating, float, np.integer, int)) and stop_criterion_vector_potential > 0.0
        cfg.stop_criterion_vector_potential = cfg.dtype(stop_criterion_vector_potential)

        # relative tolerance for convergence
        cfg.convergence_rtol = convergence_rtol

        self.cfg = cfg

        self.par = GLPar.Startup()

        self.par.red = GLPar.Reduction(self.par)

        self.mesh = GLMesh.Grid()

        self.vars = GLVars.Vars(self.par, self.mesh)

        self.params = GLVars.Params(self.mesh, self.vars)

        self.observables = GLObs.Observables(self.par, self.mesh, self.vars,
                self.params)

        self.solve = GLSolvers.Solvers(self.par, self.mesh, self.vars,
                self.params, self.observables)

        self.vortex_detector = GLObs.VortexDetector(self.vars, self.params,
                self.solve)


    def __del__(self):
        pass


# --------------------------------------------------------------------------------

    # Used in tests, so including it here
    def flatten_a_array(self, a):
        return np.reshape(a.T, self.cfg.Na)


    def unflatten_a_array(self, a):
        return np.reshape(a, (self.cfg.Nya, self.cfg.Nxa)).T


    def flatten_b_array(self, b):
        return np.reshape(b.T, self.cfg.Nb)


    def unflatten_b_array(self, b):
        return np.reshape(b, (self.cfg.Nyb, self.cfg.Nxb)).T


    def flatten_array(self, psi):
        return np.reshape(psi.T, self.cfg.N)


    def unflatten_array(self, psi):
        return np.reshape(psi, (self.cfg.Ny, self.cfg.Nx)).T





# Copyright (c) Microsoft Corporation.
# Licensed under the MIT license.

import numpy as np
from pycuda import gpuarray

import svirl.config as cfg
from svirl.parallel.utils import Utils
from svirl.storage import GArray 

# TODO: rewind phases not supported yet

class TD(object):
    """This class contains methods that solve the order parameter 
    and vector potential.
    """

    def __init__(self, par, mesh, _vars, params, observables):
        self.par = par
        self.mesh = mesh
        self.vars = _vars
        self.params = params
        self.fixed_vortices = self.params.fixed_vortices
        self.observables = observables

        self.solveA = self.params.solveA

        self.__order_parameter_phase_lock_krnl = self.par.get_function('order_parameter_phase_lock')
        self.__iterate_order_parameter_jacobi_step_krnl = self.par.get_function('iterate_order_parameter_jacobi_step')
        self.__iterate_vector_potential_jacobi_step_krnl = self.par.get_function('iterate_vector_potential_jacobi_step')

        self.__xpy_r_krnl = self.par.get_function('xpy_r')
        self.__xmy_r_krnl = self.par.get_function('xmy_r')

        self._random_t = np.uint32(1)
        if cfg.random_seed is not None:
            self._random_t = np.uint32(cfg.random_seed)

        # Alloc the rhs arrays
        self.vars._tmp_node_var = GArray(like = self.vars._psi)

        shapes = [(cfg.Nxa, cfg.Nya), (cfg.Nxb, cfg.Nyb)]
        self.vars._tmp_edge_var = GArray(shape = shapes, dtype = cfg.dtype)

        A_size = self.vars.vector_potential_h().size

        self.__gab_next = gpuarray.zeros(A_size, dtype = cfg.dtype)
        self.__gpsi_next = gpuarray.empty_like(self.vars.order_parameter_h())
        self.__gr2_max = gpuarray.zeros(1, dtype = np.int32)

        # Adjust stopping criteria according to precision
        if cfg.dtype is np.float32:
            if cfg.stop_criterion_order_parameter < 1e-6:
                cfg.stop_criterion_order_parameter = 1e-6

            if cfg.stop_criterion_vector_potential < 1e-6:
                cfg.stop_criterion_vector_potential = 1e-6
        else:
            if cfg.stop_criterion_order_parameter < 1e-12:
                cfg.stop_criterion_order_parameter = 1e-12

            if cfg.stop_criterion_vector_potential < 1e-12:
                cfg.stop_criterion_vector_potential = 1e-12

        cfg.stop_criterion_order_parameter = cfg.dtype(cfg.stop_criterion_order_parameter)
        cfg.stop_criterion_vector_potential = cfg.dtype(cfg.stop_criterion_vector_potential)


    def __del__(self):
        if hasattr(self, '__gr2_max')   and self.__gr2_max   is not None:  self.__gr2_max.gpudata.free()
        if hasattr(self, '__gpsi_next') and self.__gpsi_next is not None:  self.__gpsi_next.gpudata.free()
        if hasattr(self, '__gab_next')  and self.__gab_next  is not None:  self.__gab_next.gpudata.free()


    def _set_iterator_options(self, iterator_type, Nt=None, dt=None, T=None, mandatory_definition=True):
        assert iterator_type in ['order_parameter', 'vector_potential']
        
        # if iterator_type == 'order_parameter':
        #     if Nt is None:  Nt = self.Nt
        #     if dt is None:  dt = self.dt
        #     if T  is None:  T  = self.T
        # elif iterator_type == 'vector_potential':
        #     if Nt is None:  Nt = self.NtA
        #     if dt is None:  dt = self.dtA
        #     if T  is None:  T  = self.TA
        
        if   Nt is not None and T is not None and dt is None:
            dt = float(T)/Nt
        elif Nt is not None and T is None and dt is not None:
            T = float(dt)*Nt
        elif Nt is None and T is not None and dt is not None:
            Nt = int(np.round(T/dt))
        elif Nt is not None and T is not None and dt is not None:
            assert np.isclose(T, dt*Nt)
        
        if mandatory_definition:
            assert isinstance(dt, (np.floating, float, np.integer, int)) and dt>=0.0
            assert isinstance(Nt, (np.integer, int)) and Nt>=0
        
        if iterator_type == 'order_parameter':
            self.Nt = np.int32(Nt) if Nt is not None else None
            self.dt = cfg.dtype(dt) if dt is not None else None
            self.T  = cfg.dtype(T) if T is not None else None
        elif iterator_type == 'vector_potential':
            self.NtA = np.int32(Nt) if Nt is not None else None
            self.dtA = cfg.dtype(dt) if dt is not None else None
            self.TA  = cfg.dtype(T) if T is not None else None


    def __stability_warnings_order_parameter(self):
        d = self.dt / min(cfg.dx, cfg.dy)**2
        if d > 1.0:
            if not hasattr(self, '__stability_warnings_order_parameter_shown') or self.__stability_warnings_order_parameter_shown == 0:
                print('Warning (order parameter):  dt/min(dx,dy)^2 = %g is too large' % d)
            self.__stability_warnings_order_parameter_shown = 1
        else:
            self.__stability_warnings_order_parameter_shown = 0


    def __iterate_order_parameter_gpu_ab_preprocess(self):
        gab_gabi = None

        # self.vars._vp += self.fixed_vortices._vpi; 
        if self.vars._vp is not None and self.fixed_vortices._vpi is not None:
            self.__xpy_r_krnl(
                    self.vars.vector_potential_h(), 
                    self.fixed_vortices.irregular_vector_potential_h(), 
                    np.uint32(cfg.N),
                    block = (self.par.block_size, 1, 1), 
                    grid = (self.par.grid_size, 1, 1)
                    ) 

            gab_gabi = self.vars._vp   # just a pointer
        elif self.vars._vp is not None:
            gab_gabi = self.vars._vp
        elif self.fixed_vortices._vpi is not None:
            gab_gabi = self.fixed_vortices._vpi

        if gab_gabi is not None:
            return gab_gabi.get_d_obj()

        return np.uint32(0)


    def __iterate_order_parameter_gpu_ab_postprocess(self, gab_gabi):
        # self.vars._vp -= self.fixed_vortices._vpi; 
        if self.vars._vp is not None and self.fixed_vortices._vpi is not None:
            self.__xmy_r_krnl(
                    self.vars.vector_potential_h(), 
                    self.fixed_vortices.irregular_vector_potential_h(), 
                    np.uint32(cfg.N),
                    block = (self.par.block_size, 1, 1), 
                    grid = (self.par.grid_size, 1, 1)
                    ) 


    def __iterate_order_parameter_gpu(self, gab_gabi):
        """Performs dt-iteration of self.psi on GPU"""

        # similar to gpsi_rhs = gpsi.copy(), but does not allocate new array
        Utils.copy_dtod(self.vars._tmp_node_var_h(), self.vars.order_parameter_h())
        #self.vars._tmp_node_var.need_dtoh_sync()

        for j in range(1024):
            self.__gr2_max.fill(np.int32(0))

            # TODO: prepare all cuda calls
            self.__iterate_order_parameter_jacobi_step_krnl(  
                self.dt,
                self.params.linear_coefficient_scalar_h(),
                self.params.linear_coefficient_h(),
                self.mesh.material_tiling_h(),
                gab_gabi, 

                self.vars._tmp_node_var_h(),  # psi for right-hand side; does not change during Jacobi interactions
                self.vars.order_parameter_h(),      # psi^{j} in Jacobi method
                self.__gpsi_next,                     # psi^{j+1} in Jacobi method

                self.params.order_parameter_Langevin_coefficient,
                np.uint32(j),
                self._random_t,

                cfg.stop_criterion_order_parameter,
                self.__gr2_max,
                grid  = (self.par.grid_size, 1, 1),
                block = (self.par.block_size, 1, 1), 
            )

            # swap pointers, does not change arrays
            # TODO: this is hard-wired for now since python doesn't allow
            # assignment for a function call.Sync Status not updated

            self.vars._psi._gdata, self.__gpsi_next = self.__gpsi_next, self.vars._psi._gdata
            #self.vars.order_parameter_h(), self.__gpsi_next = self.__gpsi_next, self.vars.order_parameter_h()

            # residual = max{|b-M*psi|} 
            # r2_max_norm = residual/stop_criterion 
            r2_max_norm = 1.0e-4 * cfg.dtype(self.__gr2_max.get()[0]) 

            # convergence criteria
            if r2_max_norm < 1.0: 
                break

        self._random_t += np.uint32(1)

        if self.fixed_vortices._phase_lock_ns is not None:
            block_size = 2
            grid_size = Utils.intceil(self.fixed_vortices._phase_lock_ns.size, block_size)

            self.__order_parameter_phase_lock_krnl(
                self.vars.order_parameter_h(),
                np.int32(self.fixed_vortices._phase_lock_ns.size),
                self.fixed_vortices._phase_lock_ns_h(),
                grid  = (grid_size, 1, 1),
                block = (block_size, 1, 1), 
            )

        self.vars._psi.need_dtoh_sync()


    def __iterate_order_parameter(self, dt = None, Nt = None, T = None):
        """Performs Nt dt-iterations of self.psi"""
        
        self._set_iterator_options(iterator_type = 'order_parameter', 
                dt = dt, Nt = Nt, T = T, mandatory_definition = True)
        self.__stability_warnings_order_parameter()
        gab_gabi = self.__iterate_order_parameter_gpu_ab_preprocess()

        for tau in range(self.Nt):
            self.__iterate_order_parameter_gpu(gab_gabi)
            self.__iterate_order_parameter_gpu_ab_postprocess(gab_gabi)


    def __stability_warnings_vector_potential(self):
        d = self.dtA*self.params.gl_parameter**2 / (self.params.normal_conductivity * min(cfg.dx,cfg.dy)**2)
        if d > 1.0:
            if not hasattr(self, '___stability_warning_vector_potential_0_shown') or self.___stability_warning_vector_potential_0_shown == 0:
                print('Warning (vector potential):  dt*kappa^2/(sigma*min(dx,dy)^2) = %g is too large' % d)
            self.__stability_warning_vector_potential_0_shown = 1
        else:
            self.__stability_warning_vector_potential_0_shown = 0
        
        d = self.dtA / (self.params.normal_conductivity * min(cfg.dx, cfg.dy))
        if d > 1.0:
            if not hasattr(self, '___stability_warning_vector_potential_1_shown') or self.___stability_warning_vector_potential_1_shown == 0:
                print('Warning (vector potential):  dt/(sigma*min(dx,dy)) = %g is too large' % d)
            self.__stability_warning_vector_potential_1_shown = 1
        else:
            self.__stability_warning_vector_potential_1_shown = 0


    def __iterate_vector_potential_gpu(self):
        """Performs dtA-iteration of self.a/self.b on GPU"""

        # self.gabi += self.gab; no memory allocation
        if self.fixed_vortices._vpi is not None:
            self.__xpy_r_krnl(
                 self.fixed_vortices.irregular_vector_potential_h(), 
                 self.vars.vector_potential_h(),
                 np.uint32(cfg.N),
                 block = (self.par.block_size, 1, 1), 
                 grid = (self.par.grid_size, 1, 1)
                 ) 
            gabi_gab = self.fixed_vortices.irregular_vector_potential_h()  # just a pointer
        else:
            gabi_gab = self.vars.vector_potential_h()

        # similar to gab_rhs = gab.copy(), but does not allocate new array
        Utils.copy_dtod(self.vars._tmp_edge_var_h(), self.vars.vector_potential_h())
        #self.vars._tmp_edge_var.need_dtoh_sync()

        # if self.ab_langevin_c > 1e-16:
        #     self.gab_rhs += self.ab_langevin_c*(curand(self.gab_rhs.shape, dtype=cfg.dtype) - 0.5)
        for j in range(1024):
            self.__gr2_max.fill(np.int32(0))

            self.__iterate_vector_potential_jacobi_step_krnl(
                self.dt, 

                self.params.gl_parameter_squared_h(),
                self.params._rho,
                self.params.homogeneous_external_field,

                self.mesh.material_tiling_h(),
                self.vars.order_parameter_h(),
                gabi_gab,

                self.vars._tmp_edge_var_h(),        # ab for right-hand side; does not change during Jacobi interactions
                self.vars.vector_potential_h(),     # ab^{j} in Jacobi method
                self.__gab_next,                      # ab^{j+1} in Jacobi method

                self.params.vector_potential_Langevin_coefficient,
                np.uint32(j),
                self._random_t,
                
                cfg.stop_criterion_vector_potential,
                self.__gr2_max,
                grid  = (self.par.grid_size, 1,  1),
                block = (self.par.block_size, 1, 1), 
            )

            # swap pointers, does not change arrays
            self.vars._vp._gdata, self.__gab_next = self.__gab_next, self.vars._vp._gdata
            #self.vars.vector_potential_h(), self.__gab_next = self.__gab_next, self.vars.vector_potential_h()

            # r2_max_norm = residual/stop_criterion 
            r2_max_norm = 1.0e-4 * cfg.dtype(self.__gr2_max.get()[0]) 

            # convergence criteria
            if r2_max_norm < 1.0: 
                break

        self._random_t += np.uint32(1)

        self.vars._vp.need_dtoh_sync()
        
        # self.gabi -= self.gab; no memory allocation
        if self.fixed_vortices._vpi is not None:
            self.__xmy_r_krnl(
                    self.fixed_vortices.irregular_vector_potential_h(), 
                    self.vars.vector_potential_h(), 
                    np.uint32(cfg.N),
                    block = (self.par.block_size, 1, 1), 
                    grid = (self.par.grid_size, 1, 1)
                    ) 

 
    def __iterate_vector_potential(self, dtA = None, NtA = None, TA = None):
        """Performs NtA dtA-iterations of self.a/self.b"""
        
        if not self.solveA:
            return

        self._set_iterator_options(iterator_type = 'vector_potential', 
                dt = dtA, Nt = NtA, T = TA, mandatory_definition = True)
        self.__stability_warnings_vector_potential()

        for tau in range(self.NtA):
            self.__iterate_vector_potential_gpu()


    def __iterate(self, dt = None, Nt = None, T = None):
        """Performs Nt consequent dt-iterations of self.psi and self.a/self.b"""
        
        self._set_iterator_options(iterator_type = 'order_parameter', 
                dt = dt, Nt = Nt, T = T, mandatory_definition = True)
        self._set_iterator_options(iterator_type = 'vector_potential', 
                dt = dt, Nt = Nt, T = T, mandatory_definition = True)
        
        self.__stability_warnings_order_parameter()
        if self.solveA:
            self.__stability_warnings_vector_potential()
        
        self.td_energies = []

        for tau in range(self.Nt):
            gab_gabi = self.__iterate_order_parameter_gpu_ab_preprocess()
            self.__iterate_order_parameter_gpu(gab_gabi)
            self.__iterate_order_parameter_gpu_ab_postprocess(gab_gabi)
            if self.solveA:
                self.__iterate_vector_potential_gpu()

            # if tau%1000 == 0:
            #     E0 = self.observables.free_energy # TMP
            #     self.td_energies.append(E0)
            #     print('%3.d: E = %10.10f' % (tau, E0)) # TMP


    # solve acts as a wrapper over the iterate methods
    def _solve(self, dt = None, Nt = None, T = None, eqn = None):

        if eqn == "order_parameter":
            self.__iterate_order_parameter(dt = dt, Nt = Nt, T = T) 
        elif eqn == "vector_potential":
            self.__iterate_vector_potential(dtA = dt, NtA = Nt, TA = T) 
        else:
            self.__iterate(dt = dt, Nt = Nt, T = T)




# Copyright (c) Microsoft Corporation.
# Licensed under the MIT license.

import numpy as np
from pycuda import gpuarray
import scipy.optimize

import svirl.config as cfg
from svirl.parallel.utils import Utils


class CG(object):
    """This class contains methods that use Non-linear conjugate method to 
    minimize the free energy.
    TODO: Add more (no exposed methods)
    """

    def __init__(self, par, mesh, _vars, params, observables):
        self.par = par
        self.mesh = mesh
        self.vars = _vars
        self.params = params
        self.fixed_vortices = self.params.fixed_vortices
        self.observables = observables

        self.__free_energy_jacobian_psi_krnl_krnl = self.par.get_function('free_energy_jacobian_psi')
        self.__free_energy_jacobian_A_krnl = self.par.get_function('free_energy_jacobian_A')
        self.__free_energy_conjgrad_coef_psi_krnl = self.par.get_function('free_energy_conjgrad_coef_psi')
        self.__free_energy_conjgrad_coef_krnl = self.par.get_function('free_energy_conjgrad_coef')

        # These kernels could go to Utils
        self.__xmyx_c_sum_krnl = self.par.get_function('xmyx_c_sum')
        self.__xmyx_r_sum_krnl = self.par.get_function('xmyx_r_sum')
        self.__x_mag2_c_sum_krnl = self.par.get_function('x_mag2_c_sum')
        self.__x_mag2_r_sum_krnl = self.par.get_function('x_mag2_r_sum')
        self.__axpy_r_krnl = self.par.get_function('axpy_r')
        self.__axpy_c_krnl = self.par.get_function('axpy_c')
        self.__axmy_r_krnl = self.par.get_function('axmy_r')
        self.__axmy_c_krnl = self.par.get_function('axmy_c')
        self.__divide_scalars = self.par.get_function('divide_scalars_positive')
        
        # relative error for convergence 
        self.__convergence_rtol = cfg.convergence_rtol 

        # If only Psi is minimized, there are five coeffs in the polynomial,
        # but if both Psi and A are minimized, there are 17 coeffs.
        # TODO: this needs to accomodate changes in self.params.gl_parameter
        # (self.__gC will need to be freed and alloc'd again) 
        self.__reduction_vector_length = 5
        if self.params.solveA:
            self.__reduction_vector_length = 17

        # alloc temp arrays
        self.__gdir_psi = gpuarray.zeros(int(cfg.N), dtype = cfg.dtype_complex)
        self.__gjac_psi = gpuarray.zeros_like(self.vars.order_parameter_h())
        self.__gjac_psi_prev = gpuarray.zeros_like(self.vars.order_parameter_h())

        # numerator and denominator for beta_psi and beta_A
        self._beta_psi_num = gpuarray.zeros(1, dtype = cfg.dtype)
        self._beta_psi_den = gpuarray.zeros(1, dtype = cfg.dtype)
        self._beta_psi = gpuarray.zeros(1, dtype = cfg.dtype)

        if self.params.solveA:
            A_size = self.vars.vector_potential_h().size
            self.__gdir_A = gpuarray.zeros(A_size, dtype = cfg.dtype)
            self.__gjac_A = gpuarray.zeros(A_size, dtype = cfg.dtype) 
            self.__gjac_A_prev = gpuarray.zeros(A_size, dtype = cfg.dtype) 

            self._beta_A_num = gpuarray.zeros(1, dtype = cfg.dtype)
            self._beta_A_den = gpuarray.zeros(1, dtype = cfg.dtype)
            self._beta_A = gpuarray.zeros(1, dtype = cfg.dtype)

        self.__init_free_energy_minimization()


    def __del__(self):
        self.__free_tmp_arrays()


    # If solvers are changed, these tmp arrays should be deleted by calling
    # this method 
    def __free_tmp_arrays(self):
        if hasattr(self, '__gC') and self.__gC is not None: self.__gC.gpudata.free()

        if hasattr(self, '__gdir_psi') and self.__gdir_psi is not None: self.__gdir_psi.gpudata.free()
        if hasattr(self, '__gjac_psi') and self.__gjac_psi is not None: self.__gjac_psi.gpudata.free()
        if hasattr(self, '__gjac_psi_prev') and self.__gjac_psi_prev is not None: self.__gjac_psi_prev.gpudata.free()

        if self.params.solveA:
            if hasattr(self, '__gdir_A') and self.__gdir_A is not None: self.__gdir_A.gpudata.free()
            if hasattr(self, '__gjac_A') and self.__gjac_A is not None: self.__gjac_A.gpudata.free()
            if hasattr(self, '__gjac_A_prev') and self.__gjac_A_prev is not None: self.__gjac_A_prev.gpudata.free()


    def __init_free_energy_minimization(self):

        if hasattr(self, '__gC'):
            return

        # __gC: block reduced vector of coeffs
        ncoeffs = self.__reduction_vector_length 
        length  = int(ncoeffs*self.par.grid_size) # assumes 1D decomposition
        self.__gC = gpuarray.empty(length, dtype = cfg.dtype)

        # store the summed coeff in c
        if self.params.solveA:
            self.__c = np.zeros((5, 5), dtype = cfg.dtype)  # sums of c
        else:
            self.__c = np.zeros(5, dtype = cfg.dtype)  # sums of c


    @property
    def _free_energy_jacobian_psi(self):
        """Calculates total free energy gradient with respect to complex order parameter, psi. 
        Returns: flattened gpuarray G_jacobian_psi, where
            G_jacobian_psi.real.get() contains dG/d(psi.real)
            G_jacobian_psi.imag.get() contains dG/d(psi.imag)
        """
        # NOTE: works with material tiling
        # TODO: add external vector potential
        # TODO: add phase lock gridpoints
        # TODO: make the method private

        # TODO: in the optimized version, this synchronizations should be moved to public methods 
        self.vars._psi.sync()
        self.vars._vp.sync()

        self.__gjac_psi.fill(0.0)
        
        self.__free_energy_jacobian_psi_krnl_krnl(
            self.params.gl_parameter_squared_h(),
            self.params.linear_coefficient_scalar_h(),
            self.params.linear_coefficient_h(),
            self.params.homogeneous_external_field,

            self.mesh.material_tiling_h(),
            self.vars.order_parameter_h(),
            self.params.external_irregular_vector_potential_h(),
            self.vars.vector_potential_h(),

            self.__gjac_psi,
            grid  = (self.par.grid_size,  1,  1),
            block = (self.par.block_size, 1, 1),
        )
        
        return self.__gjac_psi


    @property
    def _free_energy_jacobian_A(self):
        """Calculates total free energy gradient with respect to vector potential, A. 
        Returns: flattened gpuarray G_jacobian_A = [dG/da, dG/db]
        """
        # TODO: add material tiling
        # TODO: add external vector potential
        # TODO: add phase lock gridpoints
        # TODO: make the method private
        
        if not self.params.solveA: 
            return None
        
        self.vars._psi.sync()
        self.vars._vp.sync()
        
        self.__gjac_A.fill(0.0)
        
        self.__free_energy_jacobian_A_krnl(
            self.params.gl_parameter_squared_h(),
            self.params.homogeneous_external_field,

            self.mesh.material_tiling_h(),
            self.vars.order_parameter_h(),
            self.params.external_irregular_vector_potential_h(),
            self.vars.vector_potential_h(),

            self.__gjac_A,
            grid  = (self.par.grid_size,  1,  1),
            block = (self.par.block_size, 1, 1),
        )
        
        return self.__gjac_A


    def _free_energy_conjgrad_coef_psi(self, __gdir_psi):
        """Calculates coefs in G(psi + alpha*dpsi) = 
        c4*alpha^4 + c3*alpha^3 + c2*alpha^2 + c1*alpha + c0
        Returns coefficients c4, c3, c2, c1, and c0."""

        # NOTE: works with material tiling
        # TODO: add external vector potential
        # TODO: add phase lock gridpoints
        # TODO: make the method private
        
        self.vars._psi.sync()
        if self.vars._vp is not None:
            self.vars._vp.sync()

        self.__free_energy_conjgrad_coef_psi_krnl(
            self.params.gl_parameter_squared_h(),
            self.params.linear_coefficient_scalar_h(),
            self.params.linear_coefficient_h(),
            self.params.homogeneous_external_field,

            self.mesh.material_tiling_h(),
            self.vars.order_parameter_h(),

            __gdir_psi,
            self.params.external_irregular_vector_potential_h(),
            self.vars.vector_potential_h(),

            self.__gC,
            grid  = (self.par.grid_size,  1,  1),
            block = (self.par.block_size, 1, 1),
        )
        
        # check if second pass is required
        if self.par.grid_size > 1:
            gCr = self.par.red.gsum_v(self.__gC, self.par.grid_size, self.__reduction_vector_length)
        else:
            gCr = self.__gC.get()

        self.__c[:] = gCr[0:5]
            
        return self.__c


    def _cg_alpha_psi_min(self):
        """Minimization of: c4*alpha^4 + c3*alpha^3 + c2*alpha^2 + c1*alpha + c0,
        which means 4.0*c4*alpha**3 + 3.0*c3*alpha**2 + 2.0*c2*alpha + c1 = 0"""
        am = np.polynomial.polynomial.polyroots([self.__c[1], 2.0*self.__c[2], 3.0*self.__c[3], 4.0*self.__c[4]])
        # print(np.isclose(am.imag, 0))
        am = am[np.isclose(am.imag, 0)].real
        am = am[am >= 0]
        am = np.min(am)
        return am


    def __free_energy_minimization_psi(self, n_iter = 1000):
        """Minimizes energy with respect to order parameter"""
        # NOTE: Tests show that 
        #       - CG minimization is much faster than TD for 1-2 vortices (at least current implementation)
        #       - for ~30 of vortices CG demonstrates similar "performance" as TD
        
        # NOTE: works with material tiling
        # TODO: add external vector potential
        # TODO: add phase lock gridpoints
        
        assert not self.params.solveA

        self.vars._psi.sync()
        self.vars._vp.sync()
        
        self.cg_energies = [] # TMP
        
        #beta = 0.0  # First iteration is steepest descent, so make beta = 0.0
        
        # gpu arrays:
        # (g)dir     : search direction
        # (g)jac     : gradient
        # (g)jac_prev: gradient from previous iteration

        self.vars._alloc_free_temporary_gpu_storage('alloc')

        self.__gdir_psi.fill(0.0)
        for i in range(n_iter):

            # 1. Compute jacobians
            self.__gjac_psi = self._free_energy_jacobian_psi
            
            # 2. Compute beta
            # Polak–Ribière formula
            # TODO: consider other formulas, see e.g. https://en.wikipedia.org/wiki/Nonlinear_conjugate_gradient_method
            if i > 0:
                # d*(d-dp)/(dp*dp)
                # = d.re, d.im]*([d.re-dp.re, d.im-dp.im])/([dp.re, dp.im]*[dp.re, dp.im])
                # = (d.re*(d.re-dp.re) + d.im*(d.im-dp.im))/(dp.re*dp.re + dp.im*dp.im)
                # = (j.re*(j.re-jp.re) + j.im*(j.im-jp.im))/(jp.re*jp.re + jp.im*jp.im)
                # where j = -d

                self.__compute_beta_psi(self.__gjac_psi, self.__gjac_psi_prev)
            
            # 3. Update search direction
            self.__axmy_c_krnl(
                    self.__gdir_psi, 
                    self.__gjac_psi, 
                    self.__gdir_psi, 
                    self._beta_psi,
                    np.uint32(cfg.N),
                    block = (self.par.block_size, 1, 1),
                    grid = (self.par.grid_size, 1, 1))
            
            # 4. Compute alpha
            self._free_energy_conjgrad_coef_psi(self.__gdir_psi)
            alpha0 = self._cg_alpha_psi_min()
            #print('iter: ', i, 'F: ', c0, c1, c2, c3, c4, 'alpha: ', alpha0, 'beta: ', beta, flush=True)
            
            # 5. Update variables
            self.__axpy_c_krnl(
                    self.__gdir_psi, 
                    self.vars.order_parameter_h(),
                    self.vars.order_parameter_h(), 
                    cfg.dtype(alpha0),
                    np.uint32(cfg.N), 
                    block = (self.par.block_size, 1, 1),
                    grid = (self.par.grid_size, 1, 1))
            
            # 6. Save 
            Utils.copy_dtod(self.__gjac_psi_prev, self.__gjac_psi)
            
            E0 = self.observables.free_energy # TMP
            self.cg_energies.append(E0) # TMP
            # if i%10 == 0:
            #     print('%3.d: E = %10.10f' % (i, E0)) # TMP

            if (i > 0  and np.abs(self.cg_energies[i]/self.cg_energies[i-1] -
                1.0) < self.__convergence_rtol):
                #print('CG converged in %d iterations with residual %g ' % ( i, np.abs(self.cg_energies[i]/self.cg_energies[i-1] - 1.0)))
                break

        self.vars._psi.need_dtoh_sync()

        self.vars._alloc_free_temporary_gpu_storage('free')


    def _free_energy_conjgrad_coef(self, __gdir_psi, __gdir_A):
        """Calculates coefs in the following polynom
        G(psi + alpha_psi*dpsi, A + alpha_A*dA) = 
           c[0,0] + c[0,1]*alpha_A + c[0,2]*alpha_A**2 + c[0,3]*alpha_A**3 + c[0,4]*alpha_A**4
        + (c[1,0] + c[1,1]*alpha_A + c[1,2]*alpha_A**2 + c[1,3]*alpha_A**3 + c[1,4]*alpha_A**4) * alpha_psi 
        + (c[2,0] + c[2,1]*alpha_A + c[2,2]*alpha_A**2 + c[2,3]*alpha_A**3 + c[2,4]*alpha_A**4) * alpha_psi**2 
        +  c[3,0] * alpha_psi**3 
        +  c[4,0] * alpha_psi**4.
        This is exact 4th power polynom in dpsi and 4th order expansion with respect to dA.
        Returns coefficient matrix c."""
        
        # TODO: add external vector potential
        # TODO: add phase lock gridpoints
        # TODO: make the method private

        # TODO: in the optimized version, this synchronizations should be moved to public methods
        self.vars._psi.sync()
        self.vars._vp.sync()
        
        self.__free_energy_conjgrad_coef_krnl(
            self.params.gl_parameter_squared_h(),
            self.params.linear_coefficient_scalar_h(),
            self.params.linear_coefficient_h(),
            self.params.homogeneous_external_field,

            self.mesh.material_tiling_h(),
            self.vars.order_parameter_h(),
            __gdir_psi,

            self.params.external_irregular_vector_potential_h(),
            self.vars.vector_potential_h(),
            __gdir_A,

            self.__gC,
            grid  = (self.par.grid_size, 1,  1),
            block = (self.par.block_size, 1, 1)
        )

        # check if second pass is required
        if self.par.grid_size > 1:
            gCr = self.par.red.gsum_v(self.__gC, self.par.grid_size, self.__reduction_vector_length)
        else:
            gCr = self.__gC.get()
        
        self.__c[0, :] = gCr[0:5]
        self.__c[1, :] = gCr[5:10]
        self.__c[2, :] = gCr[10:15]
        self.__c[3, 0] = gCr[15] 
        self.__c[4, 0] = gCr[16] 
        
        return self.__c


    def _cg_alpha_min(self, alpha0=[0.0, 0.0], tol=1e-8):
        """Minimization of (alpha_psi, alpha_A)-polynom
        F(alpha_psi, alpha_A) = sum_ij c[i,j] 
                              * alpha_psi**i * alpha_A**j"""

        c = self.__c

        # jacobian polynom
        cj0 = np.polynomial.polynomial.polyder(c, axis=0)
        cj1 = np.polynomial.polynomial.polyder(c, axis=1)

        # hessian polynom
        # ch00 = np.polynomial.polynomial.polyder(cj0, axis=0)
        # ch01 = np.polynomial.polynomial.polyder(cj0, axis=1)
        # ch11 = np.polynomial.polynomial.polyder(cj1, axis=1)

        def f(alpha):
            return np.polynomial.polynomial.polyval2d(alpha[0], alpha[1], c)

        def j(alpha):
            alpha_psi, alpha_A = alpha
            return np.array(
                [np.polynomial.polynomial.polyval2d(alpha_psi, alpha_A, cj0),
                 np.polynomial.polynomial.polyval2d(alpha_psi, alpha_A, cj1)])

        # def h(alpha):
        #     alpha_psi, alpha_A = alpha
        #     h01 = np.polynomial.polynomial.polyval2d(alpha_psi, alpha_A, ch01)
        #     return np.array(
        #           [[np.polynomial.polynomial.polyval2d(alpha_psi, alpha_A, ch00), h01], 
        #            [h01, np.polynomial.polynomial.polyval2d(alpha_psi, alpha_A, ch11)]])

        r = scipy.optimize.minimize(
            f,
            x0 = np.array(alpha0),
            jac = j,
            # hess = h,
            method = 'BFGS',
            tol = tol,
        )
        
        return r.x


    def __compute_beta_psi(self, __gjac_psi, __gjac_psi_prev):
        self.__xmyx_c_sum_krnl(__gjac_psi, __gjac_psi_prev, self.vars._tmp_psi_real_h(), np.uint32(cfg.N),
                block = (self.par.block_size, 1, 1), grid = (self.par.grid_size, 1, 1)) 
        self.par.red.gsum(self.vars._tmp_psi_real_h(), ga_out = self._beta_psi_num)
        
        self.__x_mag2_c_sum_krnl(__gjac_psi_prev, self.vars._tmp_psi_real_h(), np.uint32(cfg.N), 
                block = (self.par.block_size, 1, 1), grid = (self.par.grid_size, 1, 1))
        self.par.red.gsum(self.vars._tmp_psi_real_h(), ga_out = self._beta_psi_den)

        self.__divide_scalars(self._beta_psi_num, self._beta_psi_den, self._beta_psi,
                block = (32, 1, 1), grid = (1, 1, 1))


    def __compute_beta_A(self, __gjac_A, __gjac_A_prev):
        # TODO: try Hestenes-Stiefel formula for previous step (?)
        
        self.__xmyx_r_sum_krnl(__gjac_A, __gjac_A_prev, self.vars._tmp_A_real_h(), np.uint32(cfg.Nab), 
                block = (self.par.block_size, 1, 1), grid = (self.par.grid_size_A, 1, 1))
        self.par.red.gsum(self.vars._tmp_A_real_h(), 
                N = self.par.grid_size_A, ga_out = self._beta_A_num)
        
        self.__x_mag2_r_sum_krnl(__gjac_A_prev, self.vars._tmp_A_real_h(), np.uint32(cfg.Nab), 
                block = (self.par.block_size, 1, 1), grid = (self.par.grid_size_A, 1, 1))
        self.par.red.gsum(self.vars._tmp_A_real_h(),
                N = self.par.grid_size_A, ga_out = self._beta_A_den)

        self.__divide_scalars(self._beta_A_num, self._beta_A_den, self._beta_A, 
                block = (32, 1, 1), grid = (1, 1, 1))


    def __free_energy_minimization(self, n_iter = 1000):
        """Minimizes energy with respect to order parameter and vector potential"""
        # TODO: check material tiling
        # TODO: add external vector potential
        # TODO: add phase lock gridpoints
        
        # TODO: Ideally there should be one entry for both minimzation

        self.vars._psi.sync()
        self.vars._vp.sync()
        
        self.cg_energies = [] # TMP
        
        #beta_psi = 0.0  # First iteration is steepest descent, so make beta = 0.0
        #beta_A = 0.0  # First iteration is steepest descent, so make beta = 0.0
        
        # gpu arrays:
        # (g)dir     : search direction
        # (g)jac     : gradient
        # (g)jac_prev: gradient from previous iteration

        #cuda.start_profiler()

        self.vars._alloc_free_temporary_gpu_storage('alloc')
        
        for i in range(n_iter):

            # 1. Compute jacobians
            self.__gjac_psi = self._free_energy_jacobian_psi
            self.__gjac_A = self._free_energy_jacobian_A
            
            # 2. Compute betas 
            # use Polak–Ribière formula with resetting
            # TODO: consider other formulas, see e.g. https://en.wikipedia.org/wiki/Nonlinear_conjugate_gradient_method
            if i > 0:
                self.__compute_beta_psi(self.__gjac_psi, self.__gjac_psi_prev)
                self.__compute_beta_A(self.__gjac_A, self.__gjac_A_prev)
            
            # 3. Update search directions
            self.__axmy_c_krnl(
                    self.__gdir_psi, 
                    self.__gjac_psi, 
                    self.__gdir_psi, 
                    self._beta_psi,
                    np.uint32(cfg.N),
                    block = (self.par.block_size, 1, 1),
                    grid = (self.par.grid_size, 1, 1))

            self.__axmy_r_krnl(
                    self.__gdir_A,
                    self.__gjac_A, 
                    self.__gdir_A, 
                    self._beta_A,
                    np.uint32(cfg.Nab),
                    block = (self.par.block_size, 1, 1),
                    grid = (self.par.grid_size_A, 1, 1))

            # 4. Compute alphas
            self._free_energy_conjgrad_coef(self.__gdir_psi, self.__gdir_A)
            alpha_psi, alpha_A = self._cg_alpha_min()
            #print('iter: ', i, 'c: ', self.__c, 'alpha, beta: ', alpha_psi, alpha_A, beta_psi, beta_A, flush=True)
            
            # 5. Update variables
            self.__axpy_c_krnl(self.__gdir_psi, 
                    self.vars.order_parameter_h(), 
                    self.vars.order_parameter_h(), 
                    cfg.dtype(alpha_psi), 
                    np.uint32(cfg.N), 
                    block = (self.par.block_size, 1, 1),
                    grid = (self.par.grid_size, 1, 1))

            self.__axpy_r_krnl(
                    self.__gdir_A, 
                    self.vars.vector_potential_h(), 
                    self.vars.vector_potential_h(), 
                    cfg.dtype(alpha_A), 
                    np.uint32(cfg.Nab),
                    block = (self.par.block_size, 1, 1),
                    grid = (self.par.grid_size_A, 1, 1))

            # 6. Save previous step
            Utils.copy_dtod(self.__gjac_psi_prev, self.__gjac_psi)
            Utils.copy_dtod(self.__gjac_A_prev, self.__gjac_A)
            
            E0 = self.observables.free_energy # TMP
            self.cg_energies.append(E0) # TMP
            # if i%10 == 0:
            #     print('%3.d: E = %10.10f' % (i, E0)) # TMP

            if (i > 0  and np.abs(self.cg_energies[i]/self.cg_energies[i-1] -
                1.0) < self.__convergence_rtol):
                #print('CG converged in %d iterations with residual %g ' % ( i, np.abs(self.cg_energies[i]/self.cg_energies[i-1] - 1.0)))
                break

        #cuda.stop_profiler()

        self.vars._psi.need_dtoh_sync()
        self.vars._vp.need_dtoh_sync()
        
        self.vars._alloc_free_temporary_gpu_storage('free')

    
    def _solve(self, n_iter = 1000):
        if self.params.solveA:
            self.__free_energy_minimization(n_iter)
        else:
            self.__free_energy_minimization_psi(n_iter)



# Copyright (c) Microsoft Corporation.
# Licensed under the MIT license.

import numpy as np

import svirl.config as cfg
from svirl.storage import GArray 

class FixedVortices(object):
    """This class contains methods to fix and release vortices,
    specify irregular vector potential"""
    def __init__(self, mesh, vars):

        # irregular vector potential
        self._vpi = None

        self._phase_lock_ns = None

        self._phase_lock_radius = cfg.phase_lock_radius

        self.mesh = mesh
        self.vars = vars

        if cfg.fixed_vortices_correction is None:
            cfg.fixed_vortices_correction = 'none'

        assert cfg.fixed_vortices_correction in ('none', 'cell centers', 'vertices')

        self.fixed_vortices_correction = cfg.fixed_vortices_correction
        self.fixed_vortices = cfg.fixed_vortices


    def __del__(self):
        pass


    # |psi| around isolated vortex
    def __vortex_psi_abs(self, x, y):
        r2 = x**2 + y**2
        # (i) psi(0) = 0 and (ii) |psi(r)| = 1 - 1/(2*r^2) + O(1/r^4) for r to inf
        return (1.0 - np.exp(-r2)) / (1.0 + np.exp(-r2)) 


    def order_parameter_add_vortices(self, vortices, phase=True, deep=False):
        """Adds (global) phase winding to order parameter around fixed vortex positions"""
        vx, vy, vv = self._vortices_format(vortices)
        xg, yg = self.mesh.xy_grid
        psi = self.vars._psi.get_h()

        for k in range(vx.size):
            if phase:
                psi *= np.exp(1.0j * vv[k]                                           # add phase around vx[k], vy[k]
                                   * np.arctan2(yg-vy[k], 
                                                xg-vx[k]))
            if deep:
                psi *= np.power(                                                     # add deep at vx[k], vy[k]
                                   self.__vortex_psi_abs(xg-vx[k],
                                                         yg-vy[k]), 
                                   np.abs(vv[k]))

        self.vars._psi.need_htod_sync()
        self.vars._psi.sync()
        

    def fixed_vortices_release(self):
        """Releases fixed vortices, i.e. creates natural vortices at 
        positions (fixed_vortices_x,fixed_vortices_y) by adding phase to order parameter"""

        self.vars._psi.sync()

        psi = self.vars._psi.get_h()
        psi *= np.exp(-1.0j*self.fixed_vortices_phase)

        self.vars._psi.need_htod_sync()
        self.vars._psi.sync()
        
        self.fixed_vortices = None
        self.phase_lock_radius = None


    def _vortices_format(self, vortices):
        if vortices is None:
            vortices = [[], []]
        assert isinstance(vortices, (list, tuple, dict))
        assert len(vortices) in [2, 3]
        if isinstance(vortices, (list, tuple)):
            vx, vy = vortices[0], vortices[1]
            vv = vortices[2] if len(vortices) == 3 else []
        elif isinstance(vortices, dict):
            vx, vy = vortices['x'], vortices['y']
            vv = vortices[2] if 'vorticity' in vortices else []
        if vx is None: vx = []
        if vy is None: vy = []
        if vv is None: vv = []
        vx, vy, vv = np.array(vx).flatten(), np.array(vy).flatten(), np.array(vv).flatten()
        vN = max(vx.size, vy.size, vv.size)
        if vx.size>0 and vv.size==0: vv = np.array([1])
        assert vx.size in [1, vN] and vy.size in [1, vN] and vv.size in [1, vN]
        
        fvx, fvy, fvv = [], [], []
        x, y, v = np.nan, np.nan, np.nan
        for i in range(vN):
            if i<vx.size: x = vx[i]
            if i<vy.size: y = vy[i]
            if i<vv.size: v = vv[i]
            if np.isnan(x) or np.isnan(y) or np.isnan(v):                                 # no nans
                continue
            fvx.append(x);  fvy.append(y);  fvv.append(v)
        
        return (np.array(fvx).astype(cfg.dtype), 
                np.array(fvy).astype(cfg.dtype), 
                np.array(fvv).astype(cfg.dtype))


    def _fixed_vortices_correct(self):
        # correction of vorticity; should be integer
        self.fixed_vortices_vorticity = np.round(self.fixed_vortices_vorticity)        
        if   self.fixed_vortices_correction == 'cell centers':
            # correction of vortex position; should be in centers of cells as magnetic field B[i,j]
            self.fixed_vortices_x = cfg.dx*(                                            
                                  np.round(self.fixed_vortices_x/cfg.dx + 0.5) - 0.5)
            self.fixed_vortices_y = cfg.dy*(
                                  np.round(self.fixed_vortices_y/cfg.dy + 0.5) - 0.5)
        elif self.fixed_vortices_correction == 'vertices':
            # correction of vortex position; should be in grid vertices as order parameter psi[i,j]
            self.fixed_vortices_x = cfg.dx*np.round(self.fixed_vortices_x/cfg.dx)  
            self.fixed_vortices_y = cfg.dy*np.round(self.fixed_vortices_y/cfg.dy)


    def _ab_phase(self, a, b):
        a = cfg.dx * np.r_[np.full((1, cfg.Nya), 0.0, dtype=cfg.dtype),                # reuse a and b
                            a].cumsum(axis=0)
        b = cfg.dy * np.c_[np.full((cfg.Nxb, 1), 0.0, dtype=cfg.dtype), 
                            b].cumsum(axis=1)
        return np.repeat(b[0:1,:], cfg.Nx, axis=0) + a                                   # assume angle[0,0] = 0
        # return b + np.repeat(a[:,0:1], cfg.Ny, axis=1)                                 # equivalent expression


    @property
    def irregular_vector_potential(self):
        if self._vpi is None:
            return (np.zeros((cfg.Nxa, cfg.Nya), dtype=cfg.dtype),
                    np.zeros((cfg.Nxb, cfg.Nyb), dtype=cfg.dtype))

        self._vpi.sync()
        return self._vpi.get_vec_h() 


    def irregular_vector_potential_h(self):
        if self._vpi is not None:
            return self._vpi.get_d_obj()

        return np.uintp(0)


    @property
    def fixed_vortices_phase(self):
        ai, bi = self.irregular_vector_potential
        return self._ab_phase(ai, bi)


    @property
    def fixed_vortices(self):
        return (self.fixed_vortices_x.copy(),
                self.fixed_vortices_y.copy(),
                self.fixed_vortices_vorticity.copy())

    @fixed_vortices.setter
    def fixed_vortices(self, fixed_vortices):
        
        self.fixed_vortices_x, self.fixed_vortices_y, self.fixed_vortices_vorticity = self._vortices_format(fixed_vortices)
        self._fixed_vortices_correct()
        
        if self.fixed_vortices_x.size > 0:
            if self._vpi is None:
                shapes = [(cfg.Nxa, cfg.Nya), (cfg.Nxb, cfg.Nyb)]
                self._vpi = GArray(shape = shapes, dtype = cfg.dtype)

            ai, bi = self._vpi.get_vec_h()
            ai.fill(0.0)
            bi.fill(0.0)

            xg, yg = self.mesh.xy_grid

            for k in range(self.fixed_vortices_x.size): 
                vangle = np.arctan2(yg - self.fixed_vortices_y[k], xg - self.fixed_vortices_x[k])
                vangle -= vangle[0,0]
                
                ai += self.fixed_vortices_vorticity[k] * cfg.idx * np.diff(vangle, axis=0)
                bi += self.fixed_vortices_vorticity[k] * cfg.idy * np.diff(vangle, axis=1)

            self._vpi.need_htod_sync()
            self._vpi.sync()
        else:
            if self._vpi is not None:
                self._vpi.free()
                self._vpi = None

        self.__update_phase_lock_ns()


    @property
    def phase_lock_radius(self):
        return self._phase_lock_radius


    def __set_phase_lock_radius(self, radius):
        assert radius is None or (isinstance(radius, (np.floating, float, np.integer, int)) and radius > 0.0)
        self._phase_lock_radius = radius


    #this was fixed_vortices.setter; I changed it to phase_lock_radius.
    @phase_lock_radius.setter
    def phase_lock_radius(self, radius):
        self.__set_phase_lock_radius(radius)
        self.__update_phase_lock_ns()

        # should update config too?
        #cfg.phase_lock_radius = self._phase_lock_radius


    def __update_phase_lock_ns(self):
        if self._phase_lock_ns is not None:
            self._phase_lock_ns.free()
            self._phase_lock_ns = None
        
        if self._phase_lock_radius is not None:
            xg, yg = self.mesh.xy_grid
            lock_grid = np.full((cfg.Nx, cfg.Ny), False, dtype = np.bool)
            for k in range(self.fixed_vortices_x.size):
                lock_grid = np.logical_or(lock_grid, 
                     np.square(xg - self.fixed_vortices_x[k]) + 
                     np.square(yg - self.fixed_vortices_y[k]) <= 
                                              np.square(self._phase_lock_radius))

            ns = np.where(lock_grid)[0].astype(np.int32)
            if ns.size > 0:
                self._phase_lock_ns = GArray(like = ns)


    def _phase_lock_ns_h(self):
        if self._phase_lock_ns is not None:
            return self._phase_lock_ns.get_d_obj()

        return np.uintp(0)




# Copyright (c) Microsoft Corporation.
# Licensed under the MIT license.

import numpy as np

from svirl.solvers.cg import CG
from svirl.solvers.td import TD

import svirl.config as cfg

class Solvers(object):
    """ This class contains solver interface to two solvers:
    Non-linear Conjugate Gradient (cg)
    Time-Depenedent Ginzburg-Landau (td) 
    """

    def __init__(self, par, mesh, _vars, params, observables):

        self.par = par
        self.mesh = mesh
        self.vars = _vars
        self.params = params
        self.observables = observables

        # solvers 
        self._td = None
        self._cg = None

        # "Post" operations
        self.__detected_vortices_status = False
        # False: vortices not detected, True: vortices detected
        # This is used and updated by VortexDetector class 


    def __del__(self):
        pass


    # Exposed properties and methods

    def td(self, dt = 0.1, Nt = 1000, T = None, eqn = None):
        """ Driver function to solve the time-Depenedent 
        Ginzburg-Landau PDE for the specified number of time steps 
        or until the specified time

        Parameters
        ----------
        dt: Time step size

        Nt: Number of time steps

        T : End time

        eqn: Accepts 'order_parameter' or 'vector_potential'
             and solves that equation
        """
        self.__solve_td(dt = dt, Nt = Nt, T = T, eqn = eqn)
        self.vortices_detected = False


    def cg(self, n_iter = 1000):
        """ Driver function to minimize the free energy using
        modified non-linear conjugate gradient method.

        Parameters
        ----------
        n_iter: Maximum number of iterations (default is 1000) 

        """
        self.__solve_cg(n_iter = n_iter)
        self.vortices_detected = False


    @property
    def vortices_detected(self):
        return self.__detected_vortices_status


    @vortices_detected.setter
    def vortices_detected(self, status):
        assert isinstance(status, bool) 
        self.__detected_vortices_status = status

        
    # Internal methods

    def _init_td(self):
        if self._td is None:
            self._td = TD(self.par, self.mesh, self.vars, 
                self.params, self.observables)


    def _init_cg(self):
        if self._cg is None:
            self._cg = CG(self.par, self.mesh, self.vars, 
                    self.params, self.observables)


    def __solve_td(self, dt = None, Nt = None, T = None, eqn = None):
        assert isinstance(Nt, (np.integer, int))
        assert isinstance(dt, (np.floating, float, np.integer, int))

        if eqn is not None:
            assert eqn in ['order_parameter', 'vector_potential']

        if T is not None:
            assert isinstance(T, (np.floating, float, np.integer, int))

        self._init_td()
        self._td._solve(dt = dt, Nt = Nt, T = T, eqn = eqn)


    def __solve_cg(self, n_iter):
        self._init_cg()
        self._cg._solve(n_iter)




# Copyright (c) Microsoft Corporation.
# Licensed under the MIT license.

import numpy as np
from pycuda import gpuarray
import pycuda.driver as cuda

import svirl.config as cfg
from svirl.storage import GArray

class Grid(object):
    """This class contains methods to retrieve the grid information 
    at cell centroids, horizontal and vertical edge mid-points, 
    vertices/nodes as well as material tiling. 
    """

    def __init__(self):
        self._mt = None
        self.material_tiling = cfg.material_tiling


    def __del__(self):
        pass


    def have_material_tiling(self):
        if self._mt is None:
            return False

        return True


#--- grids ---#

    @property
    def xy(self):
        """Coordinates of grid vertices"""
        return (np.linspace(0.0, cfg.Lx, num=cfg.Nx, endpoint=True, dtype=cfg.dtype), 
                np.linspace(0.0, cfg.Ly, num=cfg.Ny, endpoint=True, dtype=cfg.dtype))


    @property
    def xy_grid(self):
        """Coordinates of grid vertices, Nc-by-Ny grid"""
        x, y = self.xy
        return np.meshgrid(x, y, indexing='ij')


    @property
    def xy_a(self):
        """Coordinates of horizontal edge centers"""
        return (np.linspace(0.5*cfg.dx, cfg.Lx-0.5*cfg.dx, num=cfg.Nxa, endpoint=True, dtype=cfg.dtype), 
                np.linspace(0.0,        cfg.Ly,            num=cfg.Nya, endpoint=True, dtype=cfg.dtype))


    @property
    def xy_a_grid(self):
        """Coordinates of horizontal edge centers, Nxa-by-Nya grid"""
        x, y = self.xy_a
        return np.meshgrid(x, y, indexing='ij')


    @property
    def xy_b(self):
        """Coordinates of vertical edge centers"""
        return (np.linspace(0.0,        cfg.Lx,            num=cfg.Nxb, endpoint=True, dtype=cfg.dtype), 
                np.linspace(0.5*cfg.dy, cfg.Ly-0.5*cfg.dy, num=cfg.Nyb, endpoint=True, dtype=cfg.dtype))


    @property
    def xy_b_grid(self):
        """Coordinates of vertical edge centers, Nxb-by-Nyb grid"""
        x, y = self.xy_b
        return np.meshgrid(x, y, indexing='ij')


    @property
    def xy_c(self):
        """Coordinates of cell centers"""
        return (np.linspace(0.5*cfg.dx, cfg.Lx-0.5*cfg.dx, num=cfg.Nxc, endpoint=True, dtype=cfg.dtype), 
                np.linspace(0.5*cfg.dy, cfg.Ly-0.5*cfg.dy, num=cfg.Nyc, endpoint=True, dtype=cfg.dtype))


    @property
    def xy_c_grid(self):
        """Coordinates of cell centers, Nxc-by-Nyc grid"""
        x, y = self.xy_c
        return np.meshgrid(x, y, indexing='ij')


    @property
    def material_tiling(self):
        """Material tiling: Returns True if material is superconducting in a cell, 
        else return False
        """
        if self._mt is not None:
            return self._mt.get_h().copy()
        else:
            return np.full((cfg.Nxc, cfg.Nyc), True, dtype=np.bool)


    @material_tiling.setter
    def material_tiling(self, material_tiling):

        if callable(material_tiling):
            xg, yg = self.xy_c_grid
            mt = material_tiling(xg, yg)
        else:
            mt = material_tiling

        if mt is not None:
            assert mt.shape == (cfg.Nxc, cfg.Nyc)

            self._mt = GArray(like = mt.astype(np.bool))
            #self.set_order_parameter_to_zero_outside_material()
            #self._mt.sync()
        else: 
            if self._mt is not None:
                self._mt.free()
                self._mt = None


    def material_tiling_h(self):
        if self._mt is not None:
            return self._mt.get_d_obj()

        return np.uintp(0)


    def __in_material(self, x, y, prohibited_length=None):
        # TODO: test in_material() method
        if isinstance(x, (np.floating, float, np.integer, int)) and isinstance(y, (np.floating, float, np.integer, int)):
            xs, ys, scalar = np.array([x], dtype=self.dtype), np.array([y], dtype=self.dtype), True
        else:
            xs, ys, scalar = x, y, False
        assert xs.shape == ys.shape
        if prohibited_length is None: prohibited_length = self.dtype(0.0)
        
        in_mt = np.full_like(xs, True, dtype=np.bool)
        
        in_mt[np.logical_or.reduce((
            xs < prohibited_length, 
            xs > self.Lx-prohibited_length,
            ys < prohibited_length,
            ys > self.Ly-prohibited_length,
        ))] = False
        
        xg_c, yg_c = self.xy_c_grid
        xg_c, yg_c = self.flatten_c_array(xg_c), self.flatten_c_array(yg_c)
        
        in_mt = in_mt.reshape(-1)
        for e, (x_, y_) in enumerate(zip(xs.ravel(), ys.ravel())):
            if np.any(np.logical_and(np.square(xg_c-x_) + np.square(yg_c-y_) < np.square(prohibited_length), ~self.mt)):
                im_mt[e] = False
        in_mt = in_mt.reshape(xs.shape)
        
        return in_mt if not scalar else in_mt[0]


    def _get_material_tiling_at_nodes(self):
        mt = self._mt.get_h()

        mt_c = np.full((1, cfg.Nyc), False, dtype=np.bool)
        mt_r = np.full((cfg.Nx, 1), False, dtype=np.bool)
        
        return (np.logical_or.reduce((
            np.c_[mt_r, np.r_[mt_c, mt]],
            np.c_[mt_r, np.r_[mt, mt_c]],
            np.c_[np.r_[mt_c, mt], mt_r],
            np.c_[np.r_[mt, mt_c], mt_r]
        )))



    def interpolate_ab_array_to_c_array(self, a, b):
        cx = 0.5*(a[:, :-1] + a[:, 1:])
        cy = 0.5*(b[:-1, :] + b[1:, :])
        return cx, cy


    def interpolate_ab_array_to_c_array_abs(self, a, b):
        cx, cy = self.interpolate_ab_array_to_c_array(a, b)
        return np.sqrt(np.square(cx) + np.square(cy))




# Copyright (c) Microsoft Corporation.
# Licensed under the MIT license.

import numpy as np

import svirl.config as cfg

from svirl.parallel.utils import Utils

class Observables(object):
    """This class contains methods to compute the physical observables"""

    def __init__(self, Par, mesh, vars, params):
        self.par = Par
        self.mesh = mesh
        self.vars = vars
        self.params = params 

        self.__magnetic_field_krnl = self.par.get_function('magnetic_field')
        self.__supercurrent_density_krnl = self.par.get_function('supercurrent_density')
        self.__current_density_krnl = self.par.get_function('current_density')
        self.__free_energy_pseudodensity_krnl = self.par.get_function('free_energy_pseudodensity')


    def __del__(self):
        pass


    @property
    def superfluid_density(self):
        """Calculates local superfluid density in grid vertices"""
        self.vars._psi.sync()
        return Utils.abs2(self.vars._psi.get_h())


    @property
    def magnetic_field(self):
        """Calculates induced magnetic field in grid cells"""

        # From SJ: this seemed redundant, so I commented it out
        #if not self.params.solveA:
        #    B = np.full((self.Nx, self.Ny), self.params.H, dtype=cfg.dtype)

        self.vars._vp.sync()
        self.params._vpei.sync()
        
        self.__magnetic_field_krnl(
            self.params.external_irregular_vector_potential_h(),
            self.vars.vector_potential_h(),
            self.vars._tmp_cell_var_h(),
            grid  = (self.par.grid_size, 1, 1),
            block = (self.par.block_size, 1, 1), 
        )

        self.vars._tmp_cell_var.need_dtoh_sync()
        B = self.vars._tmp_cell_var.get_h()

        return B.copy()


    @property
    def supercurrent_density(self):
        """Calculates superconducting current density on grid edges"""
        self.vars._psi.sync()
        self.vars._vp.sync()

        if self.params.external_irregular_vector_potential_h():
            self.params._vpei.sync()

        self.__supercurrent_density_krnl(
            self.mesh.material_tiling_h(),
            self.vars.order_parameter_h(),
            self.params.external_irregular_vector_potential_h(),
            self.vars.vector_potential_h(),
            self.vars._tmp_edge_var_h(),
            grid  = (self.par.grid_size, 1, 1),
            block = (self.par.block_size, 1, 1), 
        )

        self.vars._tmp_edge_var.need_dtoh_sync()
        jsx, jsy = self.vars._tmp_edge_var.get_vec_h()

        return (jsx.copy(), jsy.copy())


    @property
    def current_density(self):
        """Calculates total current density on grid edges"""
        if not self.params.solveA:
            return self.supercurrent_density

        self.vars._vp.sync()

        if self.params.external_irregular_vector_potential_h():
            self.params._vpei.sync()
        
        self.__current_density_krnl(
            self.params.gl_parameter_squared_h(),
            self.params.homogeneous_external_field,
            self.params.external_irregular_vector_potential_h(),
            self.vars.vector_potential_h(),
            self.vars._tmp_edge_var_h(), # Re-use for jxjy
            grid  = (self.par.grid_size, 1, 1),
            block = (self.par.block_size, 1, 1), 
        )

        self.vars._tmp_edge_var.need_dtoh_sync()
        jx, jy = self.vars._tmp_edge_var.get_vec_h()
        
        return (jx.copy(), jy.copy())


    @property
    def normalcurrent_density(self):
        """Calculates normal current density on grid edges"""
        
        # TODO: consider cache current_density and supercurrent_density or run separate GPU routine for normalcurrent_density
        jx, jy = self.current_density
        jsx, jsy = self.supercurrent_density

        return (jx - jsx, jy - jsy)


    @property
    def free_energy(self):
        """Calculates total GL free energy of the system"""

        self.vars._psi.sync()
        self.vars._vp.sync()
        
        self.vars._alloc_free_temporary_gpu_storage('alloc')

        self.__free_energy_pseudodensity_krnl(
            self.params.gl_parameter_squared_h(),
            self.params.linear_coefficient_scalar_h(),
            self.params.linear_coefficient_h(),
            self.params.homogeneous_external_field,

            self.mesh.material_tiling_h(),
            self.vars.order_parameter_h(),
            self.params.external_irregular_vector_potential_h(),
            self.vars.vector_potential_h(),
            self.vars._tmp_psi_real_h(), 
            grid  = (self.par.grid_size, 1, 1),
            block = (self.par.block_size, 1, 1),
        )
        G = self.par.red.gsum(self.vars._tmp_psi_real.get_d_obj())
            
        return G





# Copyright (c) Microsoft Corporation.
# Licensed under the MIT license.

import numpy as np

from pycuda import gpuarray
import pycuda.driver as cuda

from warnings import warn

class GArray(object):
    """Returns a GArray object upon allocating memory on either
    host, device, or host and device. 

    Allocations on host and device are done using numpy and PyCUDA's
    gpuarray respectively. Allocations on multiple devices are not 
    supported yet.

    Assumes:
      o cuda context for this device has already been created.
      o GPUArrays are 1D while numpy arrays can be multi-dimensional.
         if multi-dimensional arrays are given as inputs, the memory
         allocation will correspond to their flattened version

    Parameters
    ----------
    nelements  : (required with shape) Accepts positive integers and specifies the number of
                 elements in the array. Required if argument `like`
                 is not specified (and ignored if specified with `like`)

    shape      : (optional) Accepts a tuple of integers that specifies the shape of
                 the array on host. The corresponding array on GPU, if 
                 requested, will be flattened. shape will be ignored if
                 specified with the parameter `like` since it will be 
                 derived from the prototype

                 Shape can be a list of tuples to indicate vector
                 storage. The corresponding array on GPU, if requested,
                 will be flattened in the same order as the list.

    dtype      : (Optional) Accepts only numpy float and complex datatypes. 
                 If not specified, np.float64 is assumed. 
                 Supported dtypes:
                  (np.float32, np.float64, np.complex64, np.complex128) 

    on         : (Optional) Accepts only one of the following class variables:
                  on_device,
                  on_host_and_device

    like       : (Optional) Accepts a numpy, pycuda's gpuarray, or a GArray object 
                 as a prototype and creates a new object inheriting the
                 properties of the prototype.

                 If prototype is GArray with allocations on both host 
                 and device, make sure the data is synced. In this case,
                 the data will be fetched from the device.

    like_init  : (Optional) Accepts a boolean. If this is set to True (default), 
                 the allocated object will be initialized to the 
                 prototype's value.

                 This property is not inherited if another GArray
                 object is passed using `like` object and will use 
                 the default value (GArray.on_host_and_device)

                 When this is set to True, the data is copied from 
                 host array, and only if is not available, data from 
                 device array is copied. If the prototype is a GArray 
                 and allocated on both host and device, make sure that
                 the data is synced before using this option.
                           
    name       : (Optional) Accepts a string as a name for the array. 
    


    Examples:
       1. Create a new array on host and device based on a host array, 
          and init the array (default)
             gabei = GArray(like = abei, on = GArray.on_device)
       2. Create a new array with size and dtype of the array as input
             psi   = GArray(N, dtype = np.complex128)
       3. Create a new array on host based on a device array, 
          and don't init the array
             a     = GArray(like = garray, like_init = False, 
                              on = GArray.on_host_and_device)


    """

    # Memory allocations can be done only on:
    on_device           = 'd'
    on_host_and_device  = 'hd'

    def __init__(self,
            nelements = None,
            dtype = np.float64,
            shape = None, 
            on = None,
            like = None,
            like_init = True, 
            name = 'array-0',
            suppress_warnings = False
            ):


        self.__suppress_warnings = suppress_warnings

        ## Sanity checks

        if on is not None:
            supported_ons = (self.on_device, self.on_host_and_device)
            if on not in supported_ons:
                raise ValueError("\n The value for `on` should be one of \
                        these: on_device, on_host_and_device")

        if like is not None:
            supported_protytypes = (np.ndarray, gpuarray.GPUArray, GArray)

            if not isinstance(like, tuple(supported_protytypes)):
                raise TypeError("\n `like` supports only the following \
                        prototypes: \n", supported_protytypes)

            if shape is not None:
                self.__warn("\n Shape argument will be ignored since \
                        a prototype is given.")

        if like is None:
            if nelements is None and shape is None:
                raise ValueError("\n Specify size using `nelements` \
                        options or specify the `shape` of the array")

            if nelements is not None and shape is not None:
                self.__warn("\n Both size and shape of the array \
                        specified. Using shape as the input.")
                nelements = None # Note that this will be updated later

        supported_types = (np.float32, np.float64, np.complex64, np.complex128) 
        if dtype not in supported_types:
            raise TypeError("\n `dtype` supports only the \
                    following types: \n", supported_types)

        # should we assert that if shape is a list, 
        # its elements should be a tuple?

        ## Fill attributes

        self._data  = None  # host array
        self._gdata = None  # device array
        self._data_v = None # pointers to host array for vector storage

        # NOTE: tested vector storage impl only for "host and device"

        self._nelements = None
        self._nelements_v = None

        # sync_status holds synchronization state between host and device:
        #    -1: Data from host need to be copied to device
        #     0: Data from device and host are synchronized
        #     1: Data from device need to be copied to host
        self.__sync_status = None

        # ndim = 1: scalar storage, 2: vector storage
        self._ndim = 1
        if shape is not None and isinstance(shape, list):
            self._ndim = len(shape)
            self._nelements_v = np.zeros(self._ndim, dtype=np.uint32)

        # 3D not supported yet
        assert self._ndim > 0 and self._ndim < 3

        self._shape = shape
        self._dtype = dtype
        self._like = like

        self.name = str(name)

        # Prototype given: Derive attributes from it
        if like is not None:
            if on is None:
                on = self.on_host_and_device

            self._nelements = like.size

            self._shape = like.shape
            self._dtype = like.dtype

        # No prototype: Fill attributes from input 
        if like is None:
            like_init = False

            if on is None:
                on = self.on_host_and_device

            # Shape provided: set nelements (_v) based on shape
            if shape is not None:
                self._shape = shape
                if self._ndim == 1:
                    self._nelements = np.prod(shape)
                else:
                    for idim in range(self._ndim):
                        self._nelements_v[idim] = np.prod(shape[idim])

                    self._nelements = np.sum(self._nelements_v)
                    self._nelements_v = np.cumsum(self._nelements_v)

            # nelements provided: set shape based on nelements
            if nelements is not None:
                assert self._ndim == 1
                self._shape = (nelements, 1)
                self._nelements = nelements

        self._on = on
        self._like_init = like_init

        if self.__on_host_and_device():
            self.synced()

        if self._shape is None:
            raise ValueError("Shape is none")

        if self._nelements is None:
            raise ValueError("nelements is none")

        # Allocate memory on host and/or device
        self.__alloc_storage()

        # TODO; for now assume we have the data on host
        if self._ndim >1:
            self._data_v = np.split(self._data, self._nelements_v) 

        # Unsave the reference to prototype
        self._like = None


    def __del__(self):
        if self.__on_device() and self._gdata is not None:
            self._gdata.gpudata.free()


    ## Exposed Methods 

    @property
    def size(self):
        """
        Returns the size of the array. For multi-dimensional arrays,
        it is the product of sizes in each dimension.
        """
        return self._nelements


    @property
    def dtype(self):
        """
        Returns the datatype (dtype) of the elements in the array.
        """
        return self._dtype


    @property
    def shape(self):
        """
        Returns the shape of the array as a tuple of integers.
        """
        return self._shape


    @property
    def on(self):
        """
        Returns where the allocations for the object have been made
        and is one of the following:

        on_device ('d'),
        on_host_and_device ('hd')
        """
        return self._on


    def sync(self):
        """
        Sync the data between host and device if needed.
        Remember to update the sync status if the data is changed. 
        This can be done using the following instance methods:

        need_htod_sync(),
        need_dtoh_sync(),
        synced()

        """
        if not self.__on_host_and_device():
            return

        self.__sync()


    def need_dtoh_sync(self):
        """
        Invoke this method to indicate that the data from
        device need to be synced with data at host. 
        It does not sync, only indicates that a sync is needed
        """
        if self.__on_host_and_device():
            self.__sync_status = 1


    def need_htod_sync(self):
        """
        Invoke this method to indicate that the data from
        host need to be synced with data at device. 
        It does not sync, only indicates that a sync is needed
        """
        if self.__on_host_and_device():
            self.__sync_status = -1


    def synced(self):
        """
        Invoke this method to indicate that the data at 
        host and device is synced.
        """
        if self.__on_host_and_device():
            self.__sync_status = 0


    # Array operations:

    def __flatten_array_with_size(self, arr, size):
        if arr is not None:
            return np.reshape(arr.T, size)

        return None


    def __flatten_array(self, arr = None):
        """
        If an input array is given, flatten it to the current size of the
        array. If not given, return the flattened data on the host
        if available, else return None
        """
        if arr is None:
            return self.__flatten_array_with_size(self._data, self._nelements)
        else:
            return self.__flatten_array_with_size(arr, self._nelements)


    def __unflatten_array_with_shape(self, arr, shape):
        if arr is not None:
            return np.reshape(arr, tuple(reversed(shape))).T

        return None


    def __unflatten_array(self, use_device_data = False):
        """
        Unflatten the device/host data to the shape of the array
        corresponding to this object.

        If `use_device_data` is True, device data is fetched and 
        unflattened. Else, host data is unflattened if available, 
        otherwise None is returned.
        """
        if use_device_data:
            return self.__unflatten_array_with_shape(self._gdata.get(), self._shape)
        else:
            return self.__unflatten_array_with_shape(self._data, self._shape)


    # Getters

    def get_h(self, sync = True):
        """
        Return the host array (not a copy)
        """
        if sync is True:
            self.sync()

        # Need to define appropriate behavior
        if self._ndim > 1:
            return None

        return self.__unflatten_array()


    def get_d(self, sync = True):
        """
        Return the device array (not a copy)
        """
        if sync is True:
            self.sync()

        # Need to define appropriate behavior
        if self._ndim > 1:
            return None

        return self.__unflatten_array(use_device_data = True)


    def get_d_obj(self, sync = False):
        """
        Return the device array object. Sync is not performed by default.
        """
        if sync is True:
            self.sync()

        return self._gdata


    def get_vec_h(self, sync = True):
        """
        Return the components of vector storage on host unflattend
        """

        if sync is True:
            self.sync()

        if self._ndim > 1:
            a = self.__unflatten_array_with_shape(self._data_v[0], self._shape[0])
            b = self.__unflatten_array_with_shape(self._data_v[1], self._shape[1])
            return (a, b)

        return (None, None)

    # Setters

    def set_h(self, arr):
        """
        Copy the contents of `arr` to the host array and update
        the host to device sync status. Multi-dimensional
        arrays are flattened before copying.
        
        """
        if arr.size != self._nelements:
            warn("Size mismatch: Storage has size %d but input array has: %d" % (self._nelements, arr.size))
        else:
            np.copyto(self._data, self.__flatten_array(arr))

            self.need_htod_sync()


    def set_vec_h(self, arr_a, arr_b):
        """
        Set components of the vector storage 
        """
        assert arr_a.shape == self._shape[0]
        assert arr_b.shape == self._shape[1]

        if self._data is not None:
            data_a = self.__flatten_array_with_size(arr_a, arr_a.size)
            np.copyto(self._data_v[0], data_a)

            data_b = self.__flatten_array_with_size(arr_b, arr_b.size)
            np.copyto(self._data_v[1], data_b)

            self.need_htod_sync()


    # Memory Alloc/free
            
    def free(self):
        """
        Free the host and device memory on demand
        
        """
        if self._gdata is not None:
            self._gdata.gpudata.free()
            self._gdata = None

        if self._data is not None:
            del self._data

        if self._data_v is not None:
            del self._data_v

        self._nelements = 0
        self._shape = None
        self._dtype = None

            
    # Diagnostics/Utils

    def metadata(self):
        """
        Print object metadata like name, size, shape, on etc 
        
        """
        
        print('  Name       : ', self.name, flush = True)
        print('  size       : ', self._nelements, flush = True)
        print('  shape      : ', self._shape, flush = True)
        print('  dtype      : ', self._dtype, flush = True)
        print('  on         : ', self._on, flush = True)
        print('  sync status: ', self.__sync_status, flush = True)
        print('', flush = True)


    ### Private members


    def __on_device(self):
        return (self._on == self.on_device 
                or self._on == self.on_host_and_device)


    def __on_host_and_device(self):
        return (self._on == self.on_host_and_device)


    def __is_prototype_GArray(self):
        return isinstance(self._like, GArray)


    def __is_prototype_ndarray(self):
        return isinstance(self._like, np.ndarray)


    def __is_prototype_GPUArray(self):
        return isinstance(self._like, gpuarray.GPUArray)


    #  ------------------------------------------
    # | Prototype  | Prototype On | Requested On |
    #  ------------------------------------------
    #   GArray       HD, D           HD, D  <== Prototype host/device 
    #   ndarray        N/A           HD, D      data should be synced     
    #   GPUArray       N/A           HD, D
    #  ------------------------------------------


    # Allocations similar to the specified prototype + initialization
    def __alloc_like_init(self):

        # If prototype is GArray, use get_d_obj() to access the data 
        def alloc_from_GArray():
            self._gdata = self._like.get_d_obj().copy()
            if self.__on_host_and_device():
                self._data = self._gdata.get()
                self.synced()

        def alloc_from_ndarray():
            if self.__on_host_and_device():
                self._data = self.__flatten_array(self._like).copy()
                self._gdata = gpuarray.empty(self._like.size, dtype = self._dtype)
                cuda.memcpy_htod(self._gdata.gpudata, self._data)
                self.synced()
            elif self.__on_device():
                self._gdata = gpuarray.empty(self._like.size, dtype = self._dtype)
                flattened_array = self.__flatten_array(self._like)
                cuda.memcpy_htod(self._gdata.gpudata, flattened_array)

        def alloc_from_GPUArray():
            self._gdata = self._like.copy()
            if self.__on_host_and_device():
                self._data = self._gdata.get()
                self.synced()


        if self.__is_prototype_GArray():
            alloc_from_GArray()
        elif self.__is_prototype_ndarray():
            alloc_from_ndarray()
        elif self.__is_prototype_GPUArray():
            alloc_from_GPUArray()
        else:
            warn('Unsupported prototype!')


    def __alloc_default(self):
        if self.__on_device():
            self._gdata = gpuarray.empty(self._nelements, dtype = self._dtype)

        if self.__on_host_and_device(): 
            self._data = np.zeros(self._nelements, dtype = self._dtype)
            self._gdata = gpuarray.empty(self._nelements, dtype = self._dtype)
            self.synced()


    def __alloc_storage(self):
        if self._like_init is True:
            self.__alloc_like_init()
        else:
            self.__alloc_default()

    
    def __sync(self):
        if self.__need_dtoh_sync():       
            cuda.memcpy_dtoh(self._data, self._gdata.gpudata)
        elif self.__need_htod_sync(): 
            cuda.memcpy_htod(self._gdata.gpudata, self._data)

        self.synced()


    def __need_htod_sync(self):
        return (self.__sync_status is not None and self.__sync_status < 0)


    def __need_dtoh_sync(self):
        return (self.__sync_status is not None and self.__sync_status > 0)


    # Utils

    def __warn(self, msg):
        if not isinstance(msg, str):
            return

        if self.__suppress_warnings is False:
            warn(msg)




# Copyright (c) Microsoft Corporation.
# Licensed under the MIT license.

import numpy as np

import svirl.config as cfg
from svirl.storage import GArray 
from . import FixedVortices

class Params(object):
    """This class contains setters and getters for parameters"""

    def __init__(self, mesh, vars):
        self.mesh = mesh
        self.vars = vars
        self.fixed_vortices = FixedVortices(self.mesh, self.vars) 

        self.solveA = False 

        self.linear_coefficient = cfg.linear_coefficient    # epsilon
        self.gl_parameter = cfg.gl_parameter                # kappa
        self.normal_conductivity = cfg.normal_conductivity  # sigma

        # homogeneous external magnetic field
        self._H = cfg.dtype(0.0)
        self.homogeneous_external_field_reset = cfg.homogeneous_external_field

        # x- and y- components of external vector potential for non-homogeneous external magnetic field
        self.ae, self.be = None, None

        # external and irregular vector potential
        # it should be kept self._vpei = (self.ae, self.be) + (ai, bi)
        self._vpei = None

        # non-homogeneous external magnetic field
        self.external_field = cfg.external_field

        self.order_parameter_Langevin_coefficient = cfg.order_parameter_Langevin_coefficient

        self.vector_potential_Langevin_coefficient = cfg.vector_potential_Langevin_coefficient


    def __del__(self):
        pass


    @property
    def linear_coefficient(self):
        """ Sets/gets epsilon (linear coefficient)"""
        if self._epsilon.size == 1:
            return np.full((cfg.Nx, cfg.Ny), self._epsilon.get_h(), dtype = cfg.dtype)
        else:
            return self._epsilon.get_h() 


    @linear_coefficient.setter
    def linear_coefficient(self, linear_coefficient):
        if callable(linear_coefficient):
            xg, yg = mesh.xy_grid
            lc = linear_coefficient(xg, yg)
        else:
            lc = linear_coefficient

        if np.isscalar(lc):
            lc = lc*np.ones(1)
        else:
            assert lc.shape == (cfg.Nx, cfg.Ny)

        self._epsilon = GArray(like = lc.astype(cfg.dtype))


    def linear_coefficient_h(self):
        if self._epsilon.size != 1:
            return self._epsilon.get_d_obj()

        return np.uintp(0)


    def linear_coefficient_scalar_h(self):
        if self._epsilon.size == 1:
            return self._epsilon.get_h()

        return cfg.dtype(0.0)


    @property
    def gl_parameter(self):
        """ Sets/gets GL parameter"""
        return self._kappa


    @gl_parameter.setter
    def gl_parameter(self, gl_parameter):
        if gl_parameter is None or np.isnan(gl_parameter) or np.isinf(gl_parameter): gl_parameter = np.inf
        assert isinstance(gl_parameter, (np.floating, float, np.integer, int)) and (np.isposinf(gl_parameter) or gl_parameter > 0.0)
        self._kappa = cfg.dtype(gl_parameter)

        self.solveA = np.bool(not np.isposinf(self._kappa))


    def gl_parameter_squared_h(self):
        if self.solveA:
            return cfg.dtype(self.gl_parameter**2) 

        return cfg.dtype(-1.0)


    @property
    def normal_conductivity(self):
        """ Sets/gets normal conductivity"""
        return self._sigma


    @normal_conductivity.setter
    def normal_conductivity(self, normal_conductivity):
        assert isinstance(normal_conductivity, (np.floating, float, np.integer, int)) and normal_conductivity > 0.0
        self._sigma = cfg.dtype(normal_conductivity)
        self._rho = cfg.dtype(1.0/normal_conductivity)


    @property
    def homogeneous_external_field(self):
        """
        Sets/gets homogeneous external field and 
        does not update vector potential.
        """
        return self._H


    @homogeneous_external_field.setter
    def homogeneous_external_field(self, homogeneous_external_field):
        self._H = cfg.dtype(homogeneous_external_field)


    def _update_vector_potential(self, homogeneous_external_field, reset):
        assert isinstance(homogeneous_external_field, (np.floating, float, np.integer, int))
        if reset:
            self._H = cfg.dtype(homogeneous_external_field)

            # TODO: need a fill method in GArray
            # self.a.fill(0.0)
            # self.b.fill(0.0)

            a, b = self.vars._vp.get_vec_h()

            a.fill(0.0)
            b.fill(0.0)

            self.vars._vp.need_htod_sync()
            self.vars._vp.sync()

            delta_H = self._H
        else:
            delta_H = - self._H
            self._H = cfg.dtype(homogeneous_external_field)
            delta_H += self._H
            self.vars._vp.sync()
        
        # TODO: implement GPU version of ab initialization
        # Possible set of gauges, A = [g*y*H, (1-g)*x*H, 0] with any g, 0 <= g <= 1
        g = 0.5
        _, yg = self.mesh.xy_a_grid
        xg, _ = self.mesh.xy_b_grid

        a, b = self.vars._vp.get_vec_h()
        a -=  g        * (yg - 0.5*cfg.Ly) * delta_H
        b += (1.0 - g) * (xg - 0.5*cfg.Lx) * delta_H

        self.vars._vp.need_htod_sync()
        self.vars._vp.sync()


    def _homogeneous_external_field_delta(self, homogeneous_external_field):
        self._update_vector_potential(homogeneous_external_field, reset=False)

    homogeneous_external_field_delta = property(
        fset = _homogeneous_external_field_delta, 
        doc = """Sets homogeneous external field, H, and adds to the vector 
              potential deltaA, satisfying curl(deltaA) = deltaH, where 
              deltaH = H - Hold and Hold is homogeneous external field 
              before update.""")


    def _homogeneous_external_field_reset(self, homogeneous_external_field):
        self._update_vector_potential(homogeneous_external_field, reset=True)

    homogeneous_external_field_reset = property(
        fset = _homogeneous_external_field_reset,
        doc = """Sets homogeneous external field, H, and sets vector 
              potential, A, satisfying curl(A) = H.""")


    def _update_gvpei(self):
        """Sets self.gvpei = (self.ae, self.be) + (ai, bi).
        To be executed in self.external_vector_potential and self.fixed_vortices setters."""
        
        assert (self.ae is None) == (self.be is None)

        ai, bi = None, None
        if self.fixed_vortices is not None and self.fixed_vortices._vpi is not None:
            ai, bi = self.fixed_vortices._vpi.get_vec_h()
            assert (ai is None) == (bi is None)

        vpei = None
        if self.ae is not None:
            if ai is not None:
                vpei = (self.ae + ai, self.be + bi)
            else:
                vpei = (self.ae, self.be)
        else:
                vpei = (ai, bi)
        
        if self._vpei is not None and vpei is None:
            self._vpei.free()
            self._vpei = None
        else:
            #TODO: easier if GArray supports like for vector storage
            shapes = [vpei[0].shape, vpei[1].shape]
            self._vpei = GArray(shape = shapes, dtype = cfg.dtype)
            self._vpei.set_vec_h(vpei[0], vpei[1])
            self._vpei.sync()
        

    @property
    def external_vector_potential(self):
        """Sets/gets external vector potential."""
        assert (self.ae is None) == (self.be is None)
        
        if self.ae is not None:
            return self.ae, self.be

        return None


    @external_vector_potential.setter
    def external_vector_potential(self, external_vector_potential):
        if external_vector_potential is not None:
            Ax, Ay = external_vector_potential
            assert (Ax is None) == (Ay is None)
        else:
            Ax = None
        
        if Ax is not None:
            assert Ax.shape == (cfg.Nxa, cfg.Nya) 
            assert Ay.shape == (cfg.Nxb, cfg.Nyb)
            self.ae = Ax
            self.be = Ay
        else:
            self.ae, self.be = None, None
        
        self._update_gvpei()


    @property
    def external_irregular_vector_potential(self):
        """ Sets/gets external irregular vector potential"""
        if self._vpei is not None:
            return self._vpei.get_vec_h()

        return None


    def external_irregular_vector_potential_h(self):
        if self._vpei is not None:
            return self._vpei.get_d_obj() 

        return np.uintp(0)


    @property
    def external_field(self):
        """
        Sets/gets external (non-homogeneous) magnetic field.
        Setter accepts only a number now.
        """
        # TODO: return curl(A) for non-homogeneous external_field
        A = self.external_vector_potential
        if A is not None:
            Ax, Ay = A
            # TODO: check expression below
            return (- np.diff(Ax, axis=1) * cfg.idy
                    + np.diff(Ay, axis=0) * cfg.idx)
        else:
            return None


    @external_field.setter
    def external_field(self, external_field):
        if external_field is not None:
            # NOTE: placeholder, accepts only a number now
            # TODO: solve equation curl(Aext) = Hext(r) for nonuniform field Hext(r)
            
            # Possible set of gauges, A = [g*y*H, (1-g)*x*H, 0] with any g, 0 <= g <= 1
            g = 0.5
            _, yg = self.mesh.xy_a_grid
            xg, _ = self.mesh.xy_b_grid

            Ax = - g        * (yg - 0.5*cfg.Ly) * external_field                             
            Ay =  (1.0 - g) * (xg - 0.5*cfg.Lx) * external_field

            self.external_vector_potential = (Ax, Ay)
        else:
            self.external_vector_potential = None


    @property
    def order_parameter_Langevin_coefficient(self):
        return self._psi_langevin_c


    @order_parameter_Langevin_coefficient.setter
    def order_parameter_Langevin_coefficient(self, order_parameter_Langevin_coefficient):
        assert isinstance(order_parameter_Langevin_coefficient, (np.floating, float, np.integer, int))
        self._psi_langevin_c = cfg.dtype(order_parameter_Langevin_coefficient)


    @property
    def vector_potential_Langevin_coefficient(self):
        return self._ab_langevin_c


    @vector_potential_Langevin_coefficient.setter
    def vector_potential_Langevin_coefficient(self, vector_potential_Langevin_coefficient):
        assert isinstance(vector_potential_Langevin_coefficient, (np.floating, float, np.integer, int))
        self._ab_langevin_c = cfg.dtype(vector_potential_Langevin_coefficient)





# Copyright (c) Microsoft Corporation.
# Licensed under the MIT license.

from pycuda import gpuarray
import numpy as np
from svirl.parallel.utils import Utils

import svirl.config as cfg

class Reduction(object):
    """This class contains infrastructure to compute scalar
    and vector reductions on the GPU.
    """

    def __init__(self, Par):
        self.par = Par
        self._sum_krnl   = self.par.get_function('sum')
        self._sum_v_krnl = self.par.get_function('sum_v')

        # work array for parallel reduction
        self.gwork_s1 = None
        self.gwork_s2 = None
        self.gwork_v1 = None
        self.gwork_v2 = None


    def __del__(self):
        if hasattr(self, 'gwork_s1')  and self.gwork_s1  is not None:  self.gwork_s1.gpudata.free()
        if hasattr(self, 'gwork_s2')  and self.gwork_s2  is not None:  self.gwork_s2.gpudata.free()
        if hasattr(self, 'gwork_v1')  and self.gwork_v1  is not None:  self.gwork_v1.gpudata.free()
        if hasattr(self, 'gwork_v2')  and self.gwork_v2  is not None:  self.gwork_v2.gpudata.free()


    def __get_nearest_multiple(self, a, multiplier):
        return multiplier * Utils.intceil(a, multiplier)


    # TODO: Template the cuda code to support reduction of real and complex data types
    #         o Use ga_in.dtype instead of cfg.dtype.
    def gsum(self, ga_in, ga_out = None, N = 0, block_size = 0, use_gpuarray_sum = False):
        """Reduce a scalar.

        Args:
        ga_in : Input GPUArray of type real

        Optional Args:
        ga_out:  Store the output here in GPU memory. If
                not provided, the reduced value is brought back to host.
        N     :  No. of elements in ga_in array that should be reduced.
        block_size: Use specified block size in the reduction kernel
        use_gpuarray_sum: Use PyCUDA's reduction function (can be slow)
        """

        if ga_in is None:
            return None

        if use_gpuarray_sum:
            if ga_out is not None:
                Utils.copy_dtod(ga_out, gpuarray.sum(ga_in))
            else:
                return gpuarray.sum(ga_in).get().item()

        if N == 0:
            N = ga_in.size

        if block_size == 0:
            block_size = self.par.block_size

        block_size = int(self.__get_nearest_multiple(block_size, 32))
        grid_size  = int(Utils.intceil(N, block_size))

        if self.gwork_s1 is None:
            self.gwork_s1 = gpuarray.zeros(grid_size, dtype = cfg.dtype)
        elif self.gwork_s1.size < grid_size: 
            self.gwork_s1.gpudata.free()
            self.gwork_s1 = gpuarray.zeros(grid_size, dtype = cfg.dtype)

        if self.gwork_s2 is None:
            self.gwork_s2 = gpuarray.zeros(1, dtype = cfg.dtype) 

        niterations = 1
        if grid_size > 1:
            niterations = 2

        arr_in = ga_in

        arr_out = self.gwork_s1
        if niterations == 1 and ga_out is not None:
            arr_out = ga_out

        self._sum_krnl(arr_in, arr_out, np.uint32(N), 
                grid = (grid_size, 1, 1),
                block = (block_size, 1, 1)
                )

        # Output of previous iteration is input for the next one
        if niterations == 2:

            arr_in = arr_out
            arr_out = ga_out if ga_out is not None else self.gwork_s2

            if block_size >= grid_size:
                block_size = int(self.__get_nearest_multiple(grid_size, 32))

            self._sum_krnl(arr_in, arr_out, np.uint32(grid_size), 
                    block = (block_size, 1, 1)
                    )

        if ga_out is None:
           return arr_out.get()[0]


    # TODO: inplace reduction option
    # TODO: Add ga_out support if needed
    def gsum_v(self, ga_in, nv, ne, block_size = 0):
        """Reduce a vector.

        Args:
        ga_in : Input GPUArray of type real that has 'nv * ne' elements
        nv: No. of vectors
        ne: No. of elements per vector

        Optional Args:
        block_size: Use specified block size in the reduction kernel
        """

        if ga_in is None:
            return None

        if (nv == 1):
            return ga_in.copy()

        if block_size == 0:
            block_size = self.par.block_size

        block_size = int(self.__get_nearest_multiple(block_size, 32))
        grid_size = int(Utils.intceil(nv, block_size))

        size = int(grid_size * ne)
        if self.gwork_v1 is None:
            self.gwork_v1 = gpuarray.zeros(size, dtype = cfg.dtype)  
        elif self.gwork_v1.size < size: 
            self.gwork_v1.gpudata.free()
            self.gwork_v1 = gpuarray.zeros(size, dtype = cfg.dtype)

        if self.gwork_v2 is None:
            self.gwork_v2 = gpuarray.zeros(ne, dtype = cfg.dtype)  

        niterations = 1
        if grid_size > 1:
            niterations = 2

        self._sum_v_krnl(ga_in, self.gwork_v1, np.uint32(nv), 
                grid = (grid_size, 1, 1),
                block = (block_size, 1, 1)
                )

        # Output of previous iteration is input for the next one
        if niterations == 2:

            # reuse work array
            if self.gwork_v2.size != ne: 
                self.gwork_v2.gpudata.free()
                self.gwork_v2 = gpuarray.zeros(ne, dtype = cfg.dtype)

            self._sum_v_krnl(self.gwork_v1, self.gwork_v2, np.uint32(grid_size), 
                    block = (block_size, 1, 1))

            r = self.gwork_v2.get()
        else:
            r = self.gwork_v1.get()

        return r


    ############# Test functions #############

    # nv: no. of vectors, ne: no. elements per vector
    def test_sum_v(self, a_in, nv, ne, block_size = 256):

        assert ne is 5
        assert block_size >= 0 and block_size <= 1024

        ga_in = gpuarray.to_gpu(cfg.dtype(a_in))
        a_out = self.gsum_v(ga_in, nv, ne, block_size = block_size)

        return a_out

    def test_sum(self, a_in, N, block_size = 256):

        assert block_size >= 0 and block_size <= 1024

        ga_in  = gpuarray.to_gpu(cfg.dtype(a_in))
        a_out = self.gsum(ga_in, block_size = block_size)

        ga_in.gpudata.free()

        return a_out



# Copyright (c) Microsoft Corporation.
# Licensed under the MIT license.

import numpy as np
import pycuda.driver as cuda

class Utils(object):
    """Utilities"""

    @staticmethod
    def abs2(c):
        """Compute the abs value"""
        return np.square(c.real) + np.square(c.imag)

    @staticmethod
    def intceil(k, l):
        """ Equivalent of a ceil function"""
        return int(np.ceil(float(k)/float(l)))

    # Allow dest and src to be either GPUArray or GArray
    @staticmethod
    def copy_dtod(dest, src):
        """Copy GPUArray from src to dest (both in GPU memory)"""
        if dest is not None and src is not None:
            cuda.memcpy_dtod(dest.gpudata, src.gpudata, src.nbytes)
        else:
            print('Warning! src/dest pointer is null')




from setuptools import setup, find_packages, find_namespace_packages


def readme():
    with open('README.md') as f:
        return f.read()


# https://docs.python.org/3/distutils/setupscript.html#meta-data
class svirl_metadata():

    name = 'svirl'
    version = '1.0' # Format: major.minor.build.revision
    description = 'GPU accelerated Ginzburg-Landau equations solver'
    long_description = readme() 

    provides = []
    requires = [
        'python>=3.0',
        'pycuda>=2018.1',
        'numpy>=1.15',
        'scipy>=1.1',
        'matplotlib>=3.0',
        'PIL>=1.1.6',
        'cmocean>=1.2',
    ]

    maintainer = 'Shriram Jagannathan, Ivan Sadovskyy'
    maintainer_email = 'svirl@outlook.com'

    url = 'https://github.com/microsoft/svirl'
    license = 'MIT'

    # This is a required input
    packages = find_namespace_packages(include=['svirl*'])

    # https://pypi.org/classifiers/
    classifiers = [
        'Topic :: Scientific/Engineering',
        'Topic :: Scientific/Engineering :: Physics',
        'Topic :: Scientific/Engineering :: Mathematics',
        # 'Topic :: Software Development',
        'Intended Audience :: Science/Research',
        # 'Intended Audience :: Developers',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.7',
    ] 

    keywords = [
        'Superconductivity',
        'Vortex dynamics',
        'Ginzburg-Landau equations',
        'GL equations'
        'Time-dependent Ginzburg-Landau',
        'TDGL',
        'GL free energy minimization'
        'Non-linear conjugate gradient method',
        'Scientific computing',
        'Non-linear PDE',
        'GPU',
        'CUDA',
    ]


svirl = svirl_metadata()

# https://setuptools.readthedocs.io/en/latest/setuptools.html#developer-s-guide
setup(
    name                 = svirl.name,
    version              = svirl.version,
    description          = svirl.description,
    license              = svirl.license,
    packages             = svirl.packages,
    maintainer           = svirl.maintainer,
    maintainer_email     = svirl.maintainer_email,
    include_package_data = True,
    zip_safe             = False,
)



# Copyright (c) Microsoft Corporation.
# Licensed under the MIT license.

Nx = None
dx = None
Lx = None

Ny = None
dy = None
Ly = None

Nz = None
dz = None
Lz = None

NtA = None
dtA = None
TA = None 

material_tiling = None
order_parameter = 'random'
random_seed = None
random_level = 1.0

normal_conductivity = 1.0 
linear_coefficient = 1.0

homogeneous_external_field = 0.0
external_field = 0.0

fixed_vortices = None
fixed_vortices_correction = 'cell centers'
phase_lock_radius = None

order_parameter_Langevin_coefficient = 0.0
vector_potential_Langevin_coefficient = 0.0



# Copyright (c) Microsoft Corporation.
# Licensed under the MIT license.

import numpy as np

import svirl.config as cfg

class VortexDetector(object):
    """This class contains methods to detect vortices"""

    def __init__(self, _vars, params, solver):

        self.vars = _vars
        self.params = params
        self.fixed_vortices = self.params.fixed_vortices 
        self.solver = solver


    def __del__(self):
        pass


    def unflatten_c(self, n):
        return n % cfg.Nxc, n // cfg.Nxc


    def flatten_c(self, i, j):
        return i + cfg.Nxc*j


    @property
    def vortices(self):
        """Detects positions of vortices with precision beyond grid precision. 
        Simplified PRE 91, 023311 (2015) for 2D case"""
        # TODO: GPU version of vortex detector
        # TODO: vectorize loop in vortex detector
        
        if self.solver.vortices_detected is not True:
            self.vars._psi.sync()
            self.vars._vp.sync()

            ai, bi = None, None
            if self.fixed_vortices._vpi is not None:
                ai, bi = self.fixed_vortices._vpi.get_vec_h()

            a, b = self.vars._vp.get_vec_h()
            
            a_ai = a + ai if ai is not None else a
            b_bi = b + bi if bi is not None else b
            
            vcells = []
            psi_t = np.angle(self.vars.order_parameter)

            # looking for cells with non-zero vorticity
            for n in range(cfg.Nc):        
                i, j = self.unflatten_c(n)
                ip, jp = i+1, j+1

                t_00, t_p0 = psi_t[i, j], psi_t[ip, j]
                t_pp, t_0p = psi_t[ip, jp], psi_t[i, jp]

                v = - (0.5/np.pi)*(
                      np.mod(t_p0 - t_00 - cfg.dx*a[i , j ] + np.pi, 2.0*np.pi)
                    + np.mod(t_pp - t_p0 - cfg.dy*b[ip, j ] + np.pi, 2.0*np.pi)
                    + np.mod(t_0p - t_pp + cfg.dx*a[i , jp] + np.pi, 2.0*np.pi)
                    + np.mod(t_00 - t_0p + cfg.dy*b[i , j ] + np.pi, 2.0*np.pi)
                    - 4.0*np.pi
                    + cfg.dx*cfg.dy*self.params.homogeneous_external_field
                )

                if np.abs(v) > 0.5 and np.abs(v - np.round(v)) < 0.1:
                    vcells.append([n, v])
            
            # finds position where linearized f=0 zero between points (x1, y1) and (x2, y2) in 2D
            def find_zero(x1, y1, f1, x2, y2, f2):
                return (f2*x1 - x2*f1) / (f2 - f1), (f2*y1 - y2*f1) / (f2 - f1)
            
            # collects all points where linearized f=0 on the cell border
            def find_zero_line(x1, y1, f1, x2, y2, f2, x3, y3, f3, x4, y4, f4): 
                xy_zero = []
                if f2 * f1 < -1e-10:  xy_zero.append(find_zero(x2, y2, f2, x1, y1, f1))
                if f3 * f2 < -1e-10:  xy_zero.append(find_zero(x3, y3, f3, x2, y2, f2))
                if f4 * f3 < -1e-10:  xy_zero.append(find_zero(x4, y4, f4, x3, y3, f3))
                if f1 * f4 < -1e-10:  xy_zero.append(find_zero(x1, y1, f1, x4, y4, f4))
                return xy_zero
            
            # finds angle and intersection point of two lines, both defined by two points in 2D
            def find_intersection(l1x1, l1y1, l1x2, l1y2, l2x1, l2y1, l2x2, l2y2):  
                D = (l1x1-l1x2)*(l2y1-l2y2)-(l1y1-l1y2)*(l2x1-l2x2)

                # in [-pi, pi], cot(ph) = ((l1y1-l1y2)*(l2y1-l2y2)-(l1x1-l1x2)*(l2x1-l2x2)) / D
                ph = np.arctan2(D, (l1y1-l1y2)*(l2y1-l2y2)-(l1x1-l1x2)*(l2x1-l2x2))     

                # in [0, pi/2]
                ph = np.mod(np.abs(ph), 0.5*np.pi)  
                if np.abs(ph) > 1e-10:
                    ix = ((l1x1*l1y2-l1y1*l1x2)*(l2x1-l2x2)-(l1x1-l1x2)*(l2x1*l2y2-l2y1*l2x2)) / D
                    iy = ((l1x1*l1y2-l1y1*l1x2)*(l2y1-l2y2)-(l1y1-l1y2)*(l2x1*l2y2-l2y1*l2x2)) / D
                    return ix, iy, ph
                else:
                    return np.nan, np.nan, ph
            
            self._detected_vortices = []
            psi = self.vars.order_parameter
            for n, v in vcells:                                                           # triangulation of the vortices in marked cells
                i, j = self.unflatten_c(n)
                ip, jp = i+1, j+1

                x, y = cfg.dx*i, cfg.dy*j
                inta_00 = cfg.dx*a_ai[i, j ]
                inta_0p = cfg.dx*a_ai[i , jp]
                intb_00 = cfg.dy*b_bi[i , j]
                intb_p0 = cfg.dy*b_bi[ip, j]

                # TODO: dx*dy*H term is missed?
                tpsi_00 = psi[i , j ]
                tpsi_p0 = psi[ip, j ] * np.exp(-1j*(0.75*inta_00 + 0.25*(intb_00 + inta_0p - intb_p0)))
                tpsi_pp = psi[ip, jp] * np.exp(-1j*(0.5*(inta_00 + intb_p0) + 0.5*(intb_00 + inta_0p)))
                tpsi_0p = psi[i , jp] * np.exp(-1j*(0.75*intb_00 + 0.25*(inta_00 + intb_p0 - inta_0p)))

                # find real(psi) = 0 line
                re_xy = find_zero_line(  
                    x,        y,        np.real(tpsi_00), 
                    x+cfg.dx, y,        np.real(tpsi_p0), 
                    x+cfg.dx, y+cfg.dy, np.real(tpsi_pp), 
                    x,        y+cfg.dy, np.real(tpsi_0p)
                )

                # find imag(psi) = 0 line
                im_xy = find_zero_line(                                                   
                    x,        y,        np.imag(tpsi_00), 
                    x+cfg.dx, y,        np.imag(tpsi_p0), 
                    x+cfg.dx, y+cfg.dy, np.imag(tpsi_pp), 
                    x,        y+cfg.dy, np.imag(tpsi_0p)
                )

                if len(re_xy) == 2 and len(im_xy) == 2:
                    ix, iy, ph = find_intersection(
                            re_xy[0][0], re_xy[0][1], 
                            re_xy[1][0], re_xy[1][1], 
                            im_xy[0][0], im_xy[0][1], 
                            im_xy[1][0], im_xy[1][1])

                    if x-cfg.dx < ix < x+2.0*cfg.dx and y-cfg.dy < iy < y+2.0*cfg.dy:
                        self._detected_vortices.append([ix, iy, np.round(v)])
                # else:
                #     print('warning: n=%d, vorticity=%4.4g, cell=(%3.3g, %3.3g), len(re_xy)=%d, len(im_xy)=%d ' % (n, v, x+0.5*dx, y+0.5*dy, len(re_xy), len(im_xy)))
            
            if self._detected_vortices:
                self._detected_vortices = np.array(self._detected_vortices, dtype = cfg.dtype)
            else:
                self._detected_vortices = np.zeros((0,3), dtype = cfg.dtype)

            self.solver.vortices_detected = True
        
        return self._detected_vortices[:,0], self._detected_vortices[:,1], self._detected_vortices[:,2]

